import discord
from typing import List, Optional

from .base import DatabaseChecker


class EconomyAccount:
    def __init__(self, guild: int, member: int, database, table):
        self.guild = guild
        self.member = member
        self.database = database
        self.table = table

    def __str__(self):
        return f"<Account MEMBER={self.member}, GUILD={self.guild}>"

    @property
    def __checks(self):
        return EconomyManager.generate_checks(self.guild, self.member)

    async def currency(self):
        currency_data = await self.database.select(
            self.table, ["currency"], self.__checks
        )
        return currency_data["currency"]

    async def bank(self):
        bank_data = await self.database.select(self.table, ["bank"], self.__checks)
        return bank_data["bank"]

    async def net(self):
        return await self.bank() + await self.currency()

    async def change_currency(self, amount: int):
        currency = await self.currency()
        await self.database.update(
            self.table, {"currency": currency + amount}, self.__checks
        )

    async def change_bank(self, amount: int):
        bank_amount = await self.bank()
        await self.database.update(
            self.table, {"bank": bank_amount + amount}, self.__checks
        )


class EconomyManager(DatabaseChecker):
    def __init__(self, bot):
        super().__init__(
            [
                {
                    "guild": "snowflake",
                    "member": "snowflake",
                    "currency": "snowflake",
                    "bank": "snowflake",
                }
            ],
            ["economy"],
        )
        self.bot = bot

    @staticmethod
    def generate_checks(guild: int, member: int):
        return {"guild": guild, "member": member}

    async def create_account(self, member: discord.Member) -> None:
        self._check_database()

        await self.database.insertifnotexists(
            self.tables["economy"],
            {"guild": member.guild.id, "member": member.id, "currency": 0, "bank": 0},
            self.generate_checks(member.guild.id, member.id),
        )

    async def get_account(self, member: discord.Member) -> Optional[EconomyAccount]:
        self._check_database()

        member_data = await self.database.select(
            self.tables["economy"],
            [],
            self.generate_checks(member.guild.id, member.id),
            True,
        )

        if member_data:
            return EconomyAccount(
                member.guild.id, member.id, self.database, self.tables["economy"]
            )

        return None

    async def get_leaderboard(self, guild) -> List[EconomyAccount]:
        self._check_database()

        guild_info = await self.database.select(
            self.tables["economy"], [], {"guild": guild.id}, True
        )
        members = [
            EconomyAccount(
                member_info["guild"],
                member_info["member"],
                database=self.database,
                table=self.tables["economy"],
            )
            for member_info in sorted(
                guild_info, key=lambda x: x["bank"] + x["currency"], reverse=True
            )
        ]

        return members



__all__ = (
    "NotPlaying",
    "NotConnected",
    "NotPaused",
    "QueueEmpty",
    "AlreadyConnected",
    "AlreadyPaused",
    "QueueError",
    "SkipError",
    "UserNotConnected",
    "InvalidSkipIndex",
)


class NotPlaying(Exception):
    """Raises error when client is not playing"""


class NotConnected(Exception):
    """Raises error when client is not connected to a voice channel"""


class NotPaused(Exception):
    """Raises error when player is not paused"""


class QueueEmpty(Exception):
    """Raises error when queue is empty"""


class AlreadyConnected(Exception):
    """Raises error when client is already connected to voice"""


class AlreadyPaused(Exception):
    """Raises error when player is already paused."""


class QueueError(Exception):
    """Raises error when something is wrong with the queue"""


class SkipError(Exception):
    """Raises error when there is no song to skip to"""


class UserNotConnected(Exception):
    """Raises error when user is not connected to channel"""


class InvalidSkipIndex(Exception):
    """Raises error when the skip index is < 0"""



import asyncio
import sys
from abc import ABC, abstractmethod
from typing import Dict, Any, Optional, List, Union

import aiomysql
import aiopg
import aiosqlite
from motor import motor_asyncio

if sys.version_info >= (3, 8) and sys.platform.lower().startswith("win"):
    # Aiopg requires the event loop policy to be WindowsSelectorEventLoop, if it is not, aiopg raises an error.

    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


async def create_mysql(host, port, user, password, dbname):
    # Created this function to make sure the user has autocommit enabled.
    # we must make sure autocommit is enabled because manual commits are not working on aiomysql :)
    return await aiomysql.create_pool(
        host=host, port=port, user=user, password=password, db=dbname, autocommit=True
    )


class UnsupportedDatabase(Exception):
    """Raises error when the user tries to use an unsupported database."""


class Database(ABC):
    def __init__(self, database):
        self.database = database

    @abstractmethod
    async def close(self):
        pass

    @abstractmethod
    async def insertifnotexists(
        self, table_name: str, data: Dict[str, Any], checks: Dict[str, Any]
    ):
        pass

    @abstractmethod
    async def insert(self, table_name: str, data: Dict[str, Any]):
        pass

    @abstractmethod
    async def create_table(
        self,
        table_name: str,
        columns: Optional[Dict[str, str]] = None,
        exists: Optional[bool] = False,
    ):
        pass

    @abstractmethod
    async def update(
        self, table_name: str, data: Dict[str, Any], checks: Dict[str, Any]
    ):
        pass

    @abstractmethod
    async def updateorinsert(
        self,
        table_name: str,
        data: Dict[str, Any],
        checks: Dict[str, Any],
        insert_data: Dict[str, Any],
    ):
        pass

    @abstractmethod
    async def delete(self, table_name: str, checks: Dict[str, Any]):
        pass

    @abstractmethod
    async def select(
        self,
        table_name: str,
        keys: List[str],
        checks: Optional[Dict[str, Any]] = None,
        fetchall: Optional[bool] = False,
    ):
        pass

    @abstractmethod
    async def execute(
        self, sql_query: str, values: List[Any], fetchall: bool = True
    ) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
        pass


class _MongoDatabase(Database):
    def __str__(self):
        return f"<{self.__class__.__name__} '{self.name}'>"

    @property
    def name(self):
        return self.database.name

    async def close(self):
        self.database.client.close()

    async def insertifnotexists(self, table_name, data, checks):
        response = await self.select(table_name, [], checks, True)

        if not response:
            return await self.insert(table_name, data)

    async def insert(self, table_name, data):
        return await self.database[table_name].insert_one(data)

    async def create_table(self, table_name, _=None, exists=False):
        # create_table has an unused positional parameter to make the methods consistent between database types.

        if exists and table_name in await self.database.list_collection_names():
            return

        return await self.database.create_collection(table_name)

    async def update(self, table_name, data, checks):
        return await self.database[table_name].update_one(checks, {"$set": data})

    async def updateorinsert(self, table_name, data, checks, insert_data):
        response = await self.select(table_name, [], checks, True)

        if len(response) == 1:
            return await self.update(table_name, data, checks)

        return await self.insert(table_name, insert_data)

    async def delete(self, table_name, checks=None):
        return await self.database[table_name].delete_one(
            {} if checks is None else checks
        )

    async def select(self, table_name, keys, checks=None, fetchall=False):
        checks = {} if checks is None else checks

        if fetchall:
            fetch = self.database[table_name].find(checks)
            result = []

            async for doc in fetch:
                current_doc = {}

                for key, value in doc.items():
                    if not keys or key in keys:
                        current_doc[key] = value

                result.append(current_doc)
        else:
            fetch = await self.database[table_name].find_one(checks)
            result = {}

            if fetch is not None:
                for key, value in fetch.items():
                    if not keys or key in keys:
                        result[key] = value
            else:
                result = None

        return result

    async def execute(
        self, sql_query: str, values: List[Any], fetchall: bool = True
    ) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
        raise NotImplementedError("NoSQL databases cannot execute sql queries.")


class _SqlDatabase(Database):
    def __str__(self):
        return f"<{self.__class__.__name__}>"

    def with_commit(func):
        async def inner(self, *args, **kwargs):
            resp = await func(self, *args, **kwargs)
            if self.commit_needed:
                await self.commit()

            return resp

        return inner

    def with_cursor(func):
        async def inner(self, *args, **kwargs):
            database = await self.database.acquire() if self.pool else self.database

            if self.cursor_context:
                async with database.cursor() as cursor:
                    resp = await func(self, cursor, *args, **kwargs)

            else:
                cursor = await database.cursor()
                resp = await func(self, cursor, *args, **kwargs)
                await cursor.close()

            if self.pool:
                self.database.release(database)

            return resp

        return inner

    def __init__(self, database):
        super().__init__(database)
        self.place_holder = DATABASE_TYPES[type(database)]["placeholder"]
        self.cursor_context = DATABASE_TYPES[type(database)]["cursorcontext"]
        self.commit_needed = DATABASE_TYPES[type(database)]["commit"]
        self.quote = DATABASE_TYPES[type(database)]["quotes"]
        self.pool = DATABASE_TYPES[type(database)]["pool"]

    async def commit(self):
        if not self.pool:
            await self.database.commit()

    async def close(self):
        await self.database.close()

    async def insertifnotexists(self, table_name, data, checks):
        response = await self.select(table_name, [], checks, True)

        if not response:
            return await self.insert(table_name, data)

    @with_cursor
    @with_commit
    async def insert(self, cursor, table_name, data):
        query = f"INSERT INTO {table_name} ({', '.join(data.keys())}) VALUES ({', '.join([self.place_holder] * len(data.values()))})"
        await cursor.execute(query, list(data.values()))

    @with_cursor
    @with_commit
    async def create_table(self, cursor, table_name, columns=None, exists=False):
        query = f'CREATE TABLE {"IF NOT EXISTS" if exists else ""} {self.quote}{table_name}{self.quote} ('
        columns = [] if columns is None else columns

        for column in columns:
            query += f"\n{self.quote}{column}{self.quote} {columns[column]},"

        query = query[:-1]

        query += "\n);"
        await cursor.execute(query)

    @with_cursor
    @with_commit
    async def update(self, cursor, table_name, data, checks):
        query = f"UPDATE {table_name} SET "

        if data:
            for key in data:
                query += f"{key} = {self.place_holder}, "
            query = query[:-2]

        if checks:
            query += " WHERE "
            for check in checks:
                query += f"{check} = {self.place_holder} AND "

            query = query[:-4]

        await cursor.execute(query, list(data.values()) + list(checks.values()))

    async def updateorinsert(self, table_name, data, checks, insert_data):
        response = await self.select(table_name, [], checks, True)

        if len(response) == 1:
            return await self.update(table_name, data, checks)

        return await self.insert(table_name, insert_data)

    @with_cursor
    @with_commit
    async def delete(self, cursor, table_name, checks=None):
        checks = {} if checks is None else checks

        query = f"DELETE FROM {table_name} "

        if checks:
            query += "WHERE "
            for check in checks:
                query += f"{check} = {self.place_holder} AND "

            query = query[:-4]

        await cursor.execute(query, list(checks.values()))

    @with_cursor
    async def select(self, cursor, table_name, keys, checks=None, fetchall=False):
        checks = {} if checks is None else checks

        keys = "*" if not keys else keys
        query = f"SELECT {','.join(keys)} FROM {table_name} "

        if checks:
            query += "WHERE "
            for check in checks:
                query += f"{check} = {self.place_holder} AND "

            query = query[:-4]

        await cursor.execute(query, list(checks.values()))
        columns = [x[0] for x in cursor.description]

        result = await cursor.fetchall() if fetchall else await cursor.fetchone()
        if not result:
            return result

        return (
            [dict(zip(columns, x)) for x in result]
            if fetchall
            else dict(zip(columns, result))
        )

    @with_cursor
    @with_commit
    async def execute(
        self, cursor, sql_query: str, values: List[Any] = None, fetchall: bool = True
    ) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
        await cursor.execute(sql_query, values if values is not None else [])

        result = await cursor.fetchall() if fetchall else await cursor.fetchone()
        columns = [x[0] for x in cursor.description]
        if not result:
            return result

        return (
            [dict(zip(columns, x)) for x in result]
            if fetchall
            else dict(zip(columns, result))
        )


DATABASE_TYPES: Dict[Any, Dict[str, Any]] = {
    motor_asyncio.AsyncIOMotorDatabase: {"class": _MongoDatabase, "placeholder": None},
    aiosqlite.core.Connection: {
        "class": _SqlDatabase,
        "placeholder": "?",
        "cursorcontext": True,
        "commit": True,
        "quotes": '"',
        "pool": False,
    },
    aiopg.pool.Pool: {
        "class": _SqlDatabase,
        "placeholder": "%s",
        "cursorcontext": True,
        "commit": True,
        "quotes": '"',
        "pool": True,
    },
    aiomysql.pool.Pool: {
        "class": _SqlDatabase,
        "placeholder": "%s",
        "cursorcontext": True,
        "commit": False,
        "quotes": "`",
        "pool": True,
    },
}


DATABASES: List = [_SqlDatabase, _MongoDatabase]


class DatabaseManager:
    @staticmethod
    def connect(database):
        if type(database) not in DATABASE_TYPES:
            raise UnsupportedDatabase(
                f"Database of type {type(database)} is not supported by the database manager."
            )

        return DATABASE_TYPES[type(database)]["class"](database)



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import timedelta
from difflib import SequenceMatcher
from typing import TYPE_CHECKING, List, Union, Any, Iterable

import discord

from .base import EventManager, get_generator_response, CacheBased
from .punishments import get_relevant_punishment

if TYPE_CHECKING:
    from discord.ext import commands
    from .punishments import Punishment

__all__ = ("SpamManager", "SpamDetectionGenerator", "DefaultSpamDetectionGenerator")


class SpamDetectionGenerator(ABC):
    """
    Represents a SpamManager that filters messages to find spam.
    """

    __slots__ = ()

    @abstractmethod
    def generate(self, last_messages: List[discord.Message]) -> Union[bool, Any]:
        """
        This function is an abstract method.
        The generate function of the generator.

        :param last_messages: The last messages sent (5 is max).
        :type last_messages: List[discord.Message]
        :return: A boolean representing if the message is spam.
        :rtype: Union[bool, Any]
        """

        pass


class DefaultSpamDetectionGenerator(SpamDetectionGenerator):
    def generate(self, last_messages: List[discord.Message]) -> Union[bool, Any]:
        member = last_messages[0].author
        if member.guild_permissions.administrator:
            return False

        return (
            SpamManager.get_messages_similarity(
                [message.content for message in last_messages]
            )
            > 0.70
        )


class SpamManager(CacheBased, EventManager):
    """
    Represents a SpamManager which detects spam.
    """

    __slots__ = ("bot", "generator", "punishments", "_last_messages")

    def __init__(
        self,
        bot: commands.Bot,
        generator: SpamDetectionGenerator = None,
        wipe_cache_delay: timedelta = timedelta(seconds=5),
    ):
        CacheBased.__init__(self, bot, wipe_cache_delay)
        EventManager.__init__(self)

        self.generator = (
            generator if generator is not None else DefaultSpamDetectionGenerator
        )
        self.punishments = []

        self.bot.add_listener(self.__handle_messages, "on_message")
        self.bot.add_listener(self.__handle_messages, "on_message_edit")

    @staticmethod
    def get_messages_similarity(messages: Iterable[str]) -> float:
        """
        Gets the similarity between messages.

        :param messages: Messages to compare.
        :type messages: Iterable[str]
        :rtype: float
        :return: The similarity between messages (0-1)
        """

        results = []

        for i, message in enumerate(messages):
            for second_index, second_message in enumerate(messages):
                if i != second_index:
                    results.append(
                        SequenceMatcher(None, message, second_message).ratio()
                    )

        return sum(results) / len(results) if results else 0

    def add_punishments(self, punishments: List[Punishment]) -> None:
        self.punishments = punishments

    async def __handle_messages(self, message, edited_message=None):
        message = edited_message or message

        if not message.guild or message.author.bot:
            return

        member_last_messages = list(
            filter(lambda x: x.author == message.author, self.bot.cached_messages)
        )[-5:]

        if len(member_last_messages) <= 3 or not get_generator_response(
            self.generator, SpamDetectionGenerator, member_last_messages
        ):
            return

        # member_warnings are the number of times the member has spammed.
        member_warnings = (
            self._cache.setdefault(message.guild.id, {}).get(message.author.id, 0) + 1
        )
        self._cache[message.guild.id][message.author.id] = member_warnings

        await self.call_event("on_message_spam", member_last_messages, member_warnings)

        if punishment := get_relevant_punishment(self.punishments, member_warnings):
            await punishment.punishment_manager.punish(
                message, message.author, punishment
            )



from __future__ import annotations

import re
from abc import ABC, abstractmethod
from datetime import timedelta
from typing import TYPE_CHECKING, Union, Any, List

from .base import get_generator_response, EventManager, CacheBased
from .punishments import get_relevant_punishment

if TYPE_CHECKING:
    from discord.ext import commands
    import discord
    from .punishments import Punishment

__all__ = (
    "MessageFilter",
    "MessageResponseGenerator",
    "DefaultMessageResponseGenerator",
)


class MessageResponseGenerator(ABC):
    """
    Represents a URL response generator that filters messages and checks if they contain URLs or anything
    inappropriate.
    """

    __slots__ = ()

    @abstractmethod
    def generate(self, message: discord.Message) -> Union[bool, Any]:
        """
        This function is an abstract method.
        The generate function of the generator.

        :param message: The message to filter.
        :type message: discord.Message
        :return: A boolean representing if the message contains inappropriate content.
        :rtype: Union[bool, Any]
        """

        pass


class DefaultMessageResponseGenerator(MessageResponseGenerator):
    URL_RE = re.compile(
        r"(https?://(?:www\.|(?!www))[a-zA-Z0-9][a-zA-Z0-9-]+[a-zA-Z0-9]\.[^\s]{2,}|www\.[a-zA-Z0-9]["
        r"a-zA-Z0-9-]+[a-zA-Z0-9]\.[^\s]{2,}|https?://(?:www\.|(?!www))[a-zA-Z0-9]+\.[^\s]{2,"
        r"}|www\.[a-zA-Z0-9]+\.[^\s]{2,})"
    )
    DISCORD_INVITE_RE = re.compile(
        r"(?:(?:http|https)://)?(?:www.)?(?:disco|discord|discordapp).("
        r"?:com|gg|io|li|me|net|org)(?:/(?:invite))?/([a-z0-9-.]+)"
    )

    def generate(self, message: discord.Message) -> Union[bool, Any]:
        if message.author.guild_permissions.administrator:
            return False

        return self.URL_RE.match(message.content) or self.DISCORD_INVITE_RE.match(
            message.content
        )


class MessageFilter(EventManager, CacheBased):
    """
    Represents a discordSuperUtils message filter that filters messages and finds inappropriate content.
    """

    __slots__ = ("bot", "generator", "punishments")

    def __init__(
        self,
        bot: commands.Bot,
        generator: MessageResponseGenerator = None,
        delete_message: bool = True,
        wipe_cache_delay: timedelta = timedelta(minutes=5),
    ):
        CacheBased.__init__(self, bot, wipe_cache_delay)
        EventManager.__init__(self)

        self.generator = (
            generator if generator is not None else DefaultMessageResponseGenerator
        )
        self.delete_message = delete_message
        self.punishments = []

        self.bot.add_listener(self.__handle_messages, "on_message")
        self.bot.add_listener(self.__handle_messages, "on_message_edit")

    def add_punishments(self, punishments: List[Punishment]) -> None:
        self.punishments = punishments

    async def __handle_messages(self, message, edited_message=None):
        """
        This function is the main logic of the MessageFilter,
        Handled events: on_message, on_message_edit

        :param message: The on_message message passed by the event.
        :type message: discord.Message
        :param edited_message: The edited messages passed by the on_message_edit event, this function will use this
        incase it is not None.
        :type edited_message: discord.Message
        :return:
        """

        message = edited_message or message

        if not message.guild or message.author.bot:
            return

        if not get_generator_response(
            self.generator, MessageResponseGenerator, message
        ):
            return

        if self.delete_message:
            await message.delete()

        member_warnings = (
            self._cache.setdefault(message.guild.id, {}).get(message.author.id, 0) + 1
        )
        self._cache[message.guild.id][message.author.id] = member_warnings

        await self.call_event("on_inappropriate_message", message, member_warnings)

        if punishment := get_relevant_punishment(self.punishments, member_warnings):
            await punishment.punishment_manager.punish(
                message, message.author, punishment
            )



from typing import Optional, Union

from discord.ext import commands


def isfloat(string: str) -> bool:
    """
    This function receives a string and returns if it is a float or not
    :param string:
    :return:
    """

    try:
        float(string)
        return True
    except (ValueError, TypeError):
        return False


class TimeConvertor(commands.Converter):
    """
    Converts a given argument to an int that represents time in seconds.

    Examples
    ----------
        7d: 604800 (7 days in seconds)

        1m: 60 (1 minute in seconds)

        heyh: BadArgument ('hey' is not an int)

        100j: BadArgument ('j' is not a valid time multiplier)
    """

    async def convert(
        self, ctx: commands.Context, argument: str
    ) -> Optional[Union[int, float]]:
        time_multipliers = {
            "s": 1,
            "m": 60,
            "h": 60 * 60,
            "d": 60 * 60 * 24,
            "w": 60 * 60 * 24 * 7,
        }

        permanent = ["permanent", "perm", "0"]
        if argument.lower() in permanent:
            return 0

        if not isfloat(argument[:-1]) or argument[-1] not in time_multipliers.keys():
            raise commands.BadArgument(
                f"Invalid time argument provided, cannot convert '{argument}' to time."
            )

        return float(argument[:-1]) * time_multipliers[argument[-1]]



import asyncio
from typing import Optional, List, Dict, Union

import spotipy
from spotipy import SpotifyClientCredentials

FIELD = "items.track.name,items.track.artists(name),total"


class SpotifyClient:
    def __init__(self, client_id: str, client_secret: str, loop=None):
        self.sp = spotipy.Spotify(
            auth_manager=SpotifyClientCredentials(
                client_id=client_id, client_secret=client_secret
            )
        )
        self.loop = asyncio.get_event_loop() if loop is None else loop

    @staticmethod
    def get_type(url: str) -> Optional[str]:
        """
        This function receives a url and returns the type of the URL.
        Return examples: playlist, user, album, etc.

        :param url:
        :return:
        """
        url = url.replace("https://open.spotify.com/", "")
        return url.split("/")[-2]

    @staticmethod
    def make_title(song: Dict[str, dict]) -> str:
        """
        This function receives a song and creates a title that can be used for youtube_dl searching.
        Return example: Never Gonna Give You Up by Rick Astley

        :param song:
        :return:
        """
        artists = " ".join([artist["name"] for artist in song.get("artists")])
        return f"{song['name']} by {artists}"

    async def fetch_playlist_data(
        self, url: str, offset: int
    ) -> Dict[str, Union[int, list]]:
        """
        This function receives a URL and an offset and returns 100 tracks from that offset
        Example: Offset: 50, the URL has 160 tracks, returns tracks from 50-150 (limit is 100).

        :param url:
        :param offset:
        :return:
        """

        return await self.loop.run_in_executor(
            None,
            lambda: self.sp.playlist_items(
                playlist_id=url, fields=FIELD, offset=offset
            ),
        )

    async def fetch_full_playlist(self, url: str) -> List[Dict[str, dict]]:
        """
        This function receives a url and returns all the tracks in that URL.

        :param url:
        :return:
        """

        initial_request = await self.fetch_playlist_data(url, 0)
        total_tracks = initial_request.get("total")

        requests = list(
            await asyncio.gather(
                *(
                    self.fetch_playlist_data(url, offset)
                    for offset in range(100, total_tracks, 100)
                )
            )
        )
        requests.insert(0, initial_request)
        result_tracks = []

        for request in requests:
            result_tracks += request.get("items")

        return result_tracks

    async def get_songs(self, url: str) -> List[str]:
        """
        This function receives a URL and returns all the tracks in that URL.

        :param url:
        :return:
        """

        playlist_type = self.get_type(url)
        songs = []

        if playlist_type == "playlist":
            return [
                self.make_title(song["track"])
                for song in await self.fetch_full_playlist(url)
            ]

        if playlist_type == "track":
            songs = [
                await self.loop.run_in_executor(
                    None, lambda: self.sp.track(track_id=url)
                )
            ]

        if playlist_type == "album":
            album = await self.loop.run_in_executor(
                None, lambda: self.sp.album_tracks(album_id=url, limit=50)
            )
            songs = album.get("items")

        return [self.make_title(song) for song in songs]



from __future__ import annotations

from typing import TYPE_CHECKING

import discord

from .base import EventManager
from .punishments import Punisher

if TYPE_CHECKING:
    from discord.ext import commands
    from .punishments import Punishment


__all__ = ("KickManager",)


class KickManager(EventManager, Punisher):
    """
    A KickManager that manages kicks for guilds.
    """

    __slots__ = ("bot",)

    def __init__(self, bot: commands.Bot):
        super().__init__()
        self.bot = bot

    async def punish(
        self, ctx: commands.Context, member: discord.Member, punishment: Punishment
    ) -> None:
        try:
            await member.kick(reason=punishment.punishment_reason)
        except discord.errors.Forbidden as e:
            raise e
        else:
            await self.call_event("on_punishment", ctx, member, punishment)



from __future__ import annotations

import asyncio
import uuid
from abc import ABC, abstractmethod
from typing import List, Dict, Any, TYPE_CHECKING, Optional, Union

import discord
from discord.guild import VerificationLevel

from .base import DatabaseChecker

if TYPE_CHECKING:
    from discord.ext import commands
    from .database import Database

__all__ = (
    "DictionaryConvertible",
    "TemplateInfo",
    "TemplateRole",
    "TemplateCategory",
    "TemplateTextChannel",
    "TemplateVoiceChannel",
    "PartialTemplate",
    "Template",
    "TemplateManager",
)


class DictionaryConvertible(ABC):
    __slots__ = ()

    @classmethod
    @abstractmethod
    def from_dict(cls, convert_from: Dict[str, Any]) -> DictionaryConvertible:
        pass


class TemplateInfo(DictionaryConvertible):
    __slots__ = (
        "template_id",
        "guild",
        "afk_timeout",
        "mfa_level",
        "verification_level",
        "explict_content_filter",
        "system_channel",
        "afk_channel",
    )

    def __init__(
        self,
        template_id: str,
        guild: int,
        afk_timeout: int,
        mfa_level: int,
        verification_level: VerificationLevel,
        explict_content_filter: int,
        system_channel: int,
        afk_channel: int,
    ):
        self.template_id = template_id
        self.guild = guild
        self.afk_timeout = afk_timeout
        self.mfa_level = mfa_level
        self.verification_level = verification_level
        self.explict_content_filter = explict_content_filter
        self.system_channel = system_channel
        self.afk_channel = afk_channel

    def __repr__(self):
        return f"<TemplateInfo id={self.template_id} guild={self.guild} mfa_level={self.mfa_level}>"

    def __str__(self):
        return f"<TemplateInfo id={self.template_id} guild={self.guild}>"

    @classmethod
    def from_dict(cls, convert_from: Dict[str, Any]) -> TemplateInfo:
        convert_from["verification_level"] = VerificationLevel(
            convert_from["verification_level"] or 0
        )
        convert_from["explict_content_filter"] = discord.ContentFilter(
            convert_from["explict_content_filter"] or 0
        )

        return cls(*list(convert_from.values()))


class TemplateCategory(DictionaryConvertible):
    __slots__ = ("name", "position", "category_id", "overwrites")

    def __init__(
        self, name: str, position: int, category_id: int, overwrites: Dict[int, int]
    ):
        self.name = name
        self.position = position
        self.category_id = category_id
        self.overwrites = overwrites

    def __repr__(self):
        return f"<TemplateCategory id={self.category_id} name={self.name!r} position={self.position}>"

    def __str__(self):
        return f"<TemplateCategory name={self.name!r}>"

    @classmethod
    def from_dict(cls, convert_from: Dict[Any, Any]) -> TemplateCategory:
        return cls(*list(convert_from.values())[1:])


class TemplateTextChannel(DictionaryConvertible):
    __slots__ = (
        "name",
        "position",
        "category",
        "topic",
        "slowmode",
        "nsfw",
        "channel_id",
        "overwrites",
    )

    def __init__(
        self,
        name: str,
        position: int,
        category: int,
        topic: str,
        slowmode: int,
        nsfw: bool,
        channel_id: int,
        overwrites: Dict[int, discord.PermissionOverwrite],
    ):
        self.name = name
        self.position = position
        self.category = category
        self.topic = topic
        self.slowmode = slowmode
        self.nsfw = nsfw
        self.channel_id = channel_id
        self.overwrites = overwrites

    def __repr__(self):
        return f"<TemplateTextChannel topic={self.topic} name={self.name!r} position={self.position}>"

    def __str__(self):
        return f"<TemplateTextChannel name={self.name!r}>"

    @classmethod
    def from_dict(cls, convert_from: Dict[str, Any]) -> TemplateTextChannel:
        convert_from["nsfw"] = bool(convert_from["nsfw"])

        return cls(*list(convert_from.values())[1:])


class TemplateVoiceChannel(DictionaryConvertible):
    __slots__ = (
        "name",
        "position",
        "category",
        "bitrate",
        "user_limit",
        "channel_id",
        "overwrites",
    )

    def __init__(
        self,
        name: str,
        position: int,
        category: int,
        bitrate: int,
        user_limit: int,
        channel_id: int,
        overwrites: Dict[int, discord.PermissionOverwrite],
    ):
        self.name = name
        self.position = position
        self.category = category
        self.bitrate = bitrate
        self.user_limit = user_limit
        self.channel_id = channel_id
        self.overwrites = overwrites

    def __repr__(self):
        return f"<TemplateVoiceChannel limit={self.user_limit} name={self.name!r} position={self.position}>"

    def __str__(self):
        return f"<TemplateVoiceChannel name={self.name!r}>"

    @classmethod
    def from_dict(cls, convert_from: Dict[str, Any]) -> TemplateVoiceChannel:
        return cls(*list(convert_from.values())[1:])


class TemplateRole(DictionaryConvertible):
    __slots__ = (
        "default_role",
        "name",
        "color",
        "hoist",
        "position",
        "mentionable",
        "role_id",
        "permissions",
    )

    def __init__(
        self,
        default_role: bool,
        name: str,
        color: int,
        hoist: bool,
        position: int,
        mentionable: bool,
        role_id: int,
        permissions: discord.Permissions,
    ):
        self.default_role = default_role
        self.name = name
        self.color = color
        self.hoist = hoist
        self.position = position
        self.mentionable = mentionable
        self.role_id = role_id
        self.permissions = discord.Permissions(permissions)

    def get_raw(self):
        return {
            "name": self.name,
            "hoist": self.hoist,
            "mentionable": self.mentionable,
            "color": self.color,
            "permissions": self.permissions,
        }

    def __repr__(self):
        return f"<TemplateRole hoist={self.hoist} name={self.name!r} position={self.position} color={self.color}>"

    def __str__(self):
        return f"<TemplateRole name={self.name!r}>"

    @classmethod
    def from_dict(cls, convert_from: Dict[str, Any]) -> TemplateRole:
        convert_from["default_role"] = bool(convert_from["default_role"])
        convert_from["hoist"] = bool(convert_from["hoist"])
        convert_from["mentionable"] = bool(convert_from["mentionable"])

        return cls(*list(convert_from.values())[1:])


class PartialTemplate:
    __slots__ = ("info", "categories", "text_channels", "voice_channels", "roles")

    def __init__(
        self,
        info: TemplateInfo,
        categories: List[TemplateCategory],
        text_channels: List[TemplateTextChannel],
        voice_channels: List[TemplateVoiceChannel],
        roles: List[TemplateRole],
    ):
        self.info = info
        self.categories = categories
        self.text_channels = text_channels
        self.voice_channels = voice_channels
        self.roles = roles


class Template:
    __slots__ = (
        "database",
        "tables",
        "info",
        "categories",
        "text_channels",
        "voice_channels",
        "roles",
    )

    def __init__(
        self,
        database: Database,
        tables: Dict[str, str],
        info: TemplateInfo,
        categories: List[TemplateCategory],
        text_channels: List[TemplateTextChannel],
        voice_channels: List[TemplateVoiceChannel],
        roles: List[TemplateRole],
    ):
        self.database = database
        self.tables = tables
        self.info = info
        self.categories = categories
        self.text_channels = text_channels
        self.voice_channels = voice_channels
        self.roles = roles

    def __repr__(self):
        return f"<Template template={self.info} categories={self.categories} text_channels={self.text_channels}>"

    def __str__(self):
        return f"Template template={self.info}"

    async def delete(self) -> PartialTemplate:
        checks = {"id": self.info.template_id}

        partial = PartialTemplate(
            self.info,
            self.categories,
            self.text_channels,
            self.voice_channels,
            self.roles,
        )
        await self.database.delete(self.tables["templates"], checks)
        await self.database.delete(self.tables["categories"], checks)
        await self.database.delete(self.tables["text_channels"], checks)
        await self.database.delete(self.tables["voice_channels"], checks)
        await self.database.delete(self.tables["roles"], checks)
        await self.database.delete(self.tables["overwrites"], checks)
        return partial

    @staticmethod
    def format_overwrites(
        overwrites: Dict[int, discord.PermissionOverwrite],
        roles: Dict[int, discord.Role],
    ) -> Dict[discord.Role, discord.PermissionOverwrite]:
        result_overwrites = {}

        for role in roles:
            if role in overwrites:
                result_overwrites[roles[role]] = overwrites[role]

        return result_overwrites

    async def apply_settings(
        self,
        guild: discord.Guild,
        reason: str,
        channels: Dict[int, Union[discord.VoiceChannel, discord.TextChannel]],
    ) -> None:
        await guild.edit(
            afk_timeout=self.info.afk_timeout,
            verification_level=self.info.verification_level,
            explicit_content_filter=self.info.explict_content_filter,
            afk_channel=channels.get(self.info.afk_channel),
            system_channel=channels.get(self.info.system_channel),
            reason=reason,
        )

    async def apply_roles(
        self, guild: discord.Guild, reason: str
    ) -> Dict[int, discord.Role]:
        roles = await asyncio.gather(
            *[
                guild.default_role.edit(permissions=role.permissions, reason=reason)
                if role.default_role
                else guild.create_role(**role.get_raw())
                for role in reversed(self.roles)
            ]
        )

        return dict(
            zip(
                (x.role_id for x in reversed(self.roles)),
                (role if role else guild.default_role for role in roles),
            )
        )

    async def apply_categories(
        self, guild: discord.Guild, reason: str
    ) -> Dict[int, discord.CategoryChannel]:
        categories = await asyncio.gather(
            *[
                guild.create_category_channel(name=category.name, reason=reason)
                for category in self.categories
            ]
        )

        return dict(
            zip((category.category_id for category in self.categories), categories)
        )

    async def apply_channels(
        self,
        guild: discord.Guild,
        reason: str,
        categories: Dict[int, discord.CategoryChannel],
        roles: Dict[int, discord.Role],
    ) -> Dict[int, discord.TextChannel]:
        text_channels = await asyncio.gather(
            *[
                guild.create_text_channel(
                    name=channel.name,
                    position=channel.position,
                    slowmode_delay=channel.slowmode,
                    topic=channel.topic,
                    nsfw=channel.nsfw,
                    category=categories.get(channel.category),
                    overwrites=self.format_overwrites(channel.overwrites, roles),
                    reason=reason,
                )
                for channel in self.text_channels
            ]
        )

        return dict(
            zip((channel.channel_id for channel in self.text_channels), text_channels)
        )

    async def apply_voice_channels(
        self,
        guild: discord.Guild,
        reason: str,
        categories: Dict[int, discord.CategoryChannel],
        roles: Dict[int, discord.Role],
    ) -> Dict[int, discord.VoiceChannel]:
        voice_channels = await asyncio.gather(
            *[
                guild.create_voice_channel(
                    name=channel.name,
                    position=channel.position,
                    category=categories.get(channel.category),
                    bitrate=channel.bitrate,
                    user_limit=channel.user_limit,
                    overwrites=self.format_overwrites(channel.overwrites, roles),
                    reason=reason,
                )
                for channel in self.voice_channels
            ]
        )

        return dict(
            zip((channel.channel_id for channel in self.voice_channels), voice_channels)
        )

    async def apply(self, guild: discord.Guild) -> None:
        roles_to_delete = list(
            filter(
                lambda r: not r.managed
                and guild.default_role != r
                and guild.me.top_role.position > r.position,
                guild.roles,
            )
        )

        reason = f"Applying template {self.info.template_id}"

        await asyncio.gather(*[role.delete(reason=reason) for role in roles_to_delete])
        await asyncio.gather(
            *[channel.delete(reason=reason) for channel in guild.channels]
        )

        roles = await self.apply_roles(guild, reason)
        categories = await self.apply_categories(guild, reason)
        text_channels = await self.apply_channels(guild, reason, categories, roles)
        voice_channels = await self.apply_voice_channels(
            guild, reason, categories, roles
        )
        await self.apply_settings(guild, reason, {**text_channels, **voice_channels})

    @staticmethod
    def get_overwrite(
        overwrites: List[Dict[str, Any]], overwrite_object: int
    ) -> Dict[int, discord.PermissionOverwrite]:
        result_overwrites = {}

        for overwrite in overwrites:
            if overwrite["overwrite_object"] == overwrite_object:
                result_overwrites[
                    overwrite["overwrite_key"]
                ] = discord.PermissionOverwrite.from_pair(
                    discord.Permissions(overwrite["overwrite_pair"]),
                    discord.Permissions(overwrite["overwrite_second_pair"]),
                )

        return result_overwrites

    @classmethod
    async def get_template(
        cls, database: Database, tables: Dict[str, str], template_id: str
    ) -> Optional[Template]:
        checks = {"id": template_id}

        raw_info = await database.select(tables["templates"], [], checks)
        if not raw_info:
            return None

        raw_categories = await database.select(
            tables["categories"], [], checks, fetchall=True
        )
        raw_text_channels = await database.select(
            tables["text_channels"], [], checks, fetchall=True
        )
        raw_voice_channels = await database.select(
            tables["voice_channels"], [], checks, fetchall=True
        )
        raw_roles = await database.select(tables["roles"], [], checks, fetchall=True)
        overwrites = await database.select(
            tables["overwrites"], [], checks, fetchall=True
        )

        return cls(
            database,
            tables,
            TemplateInfo.from_dict(raw_info),
            [
                TemplateCategory.from_dict(
                    dict(
                        x,
                        **{
                            "overwrites": cls.get_overwrite(
                                overwrites, x["category_id"]
                            )
                        },
                    )
                )
                for x in raw_categories
            ],
            [
                TemplateTextChannel.from_dict(
                    dict(
                        x,
                        **{
                            "overwrites": cls.get_overwrite(overwrites, x["channel_id"])
                        },
                    )
                )
                for x in raw_text_channels
            ],
            [
                TemplateVoiceChannel.from_dict(
                    dict(
                        x,
                        **{
                            "overwrites": cls.get_overwrite(overwrites, x["channel_id"])
                        },
                    )
                )
                for x in raw_voice_channels
            ],
            [TemplateRole.from_dict(x) for x in raw_roles],
        )


class TemplateManager(DatabaseChecker):
    __slots__ = ("bot",)

    def __init__(self, bot: commands.Bot):
        super().__init__(
            [
                {
                    "id": "string",
                    "guild": "snowflake",
                    "afk_timeout": "number",
                    "mfa_level": "smallnumber",
                    "verification_level": "smallnumber",
                    "explict_content_filter": "smallnumber",
                    "system_channel": "snowflake",
                    "afk_channel": "snowflake",
                },
                {
                    "id": "string",
                    "name": "string",
                    "position": "number",
                    "category_id": "snowflake",
                },
                {
                    "id": "string",
                    "name": "string",
                    "position": "number",
                    "category": "snowflake",
                    "topic": "string",
                    "slowmode": "number",
                    "nsfw": "smallnumber",
                    "channel_id": "snowflake",
                },
                {
                    "id": "string",
                    "name": "string",
                    "position": "number",
                    "category": "snowflake",
                    "bitrate": "number",
                    "user_limit": "smallnumber",
                    "channel_id": "snowflake",
                },
                {
                    "id": "string",
                    "default_role": "smallnumber",
                    "name": "string",
                    "color": "number",
                    "hoist": "smallnumber",
                    "position": "number",
                    "mentionable": "smallnumber",
                    "role_id": "snowflake",
                    "permissions": "snowflake",
                },
                {
                    "id": "string",
                    "overwrite_object": "snowflake",
                    "overwrite_key": "snowflake",
                    "overwrite_pair": "snowflake",
                    "overwrite_second_pair": "snowflake",
                },
            ],
            [
                "templates",
                "categories",
                "text_channels",
                "voice_channels",
                "roles",
                "overwrites",
            ],
        )
        self.bot = bot

    async def get_templates(self, guild: discord.Guild = None) -> List[Template]:
        self._check_database()

        return [
            await Template.get_template(self.database, self.tables, template["id"])
            for template in await self.database.select(
                self.tables["templates"],
                ["id"],
                {"guild": guild.id} if guild else {},
                fetchall=True,
            )
        ]

    async def get_template(self, template_id: str) -> Optional[Template]:
        self._check_database()

        return await Template.get_template(self.database, self.tables, template_id)

    async def write_overwrites(
        self,
        template_id: str,
        overwrites_object: int,
        overwrites: discord.PermissionOverwrite,
    ) -> None:
        self._check_database()

        for x, y in overwrites.items():
            pairs = [pair.value for pair in y.pair()]

            await self.database.insert(
                self.tables["overwrites"],
                {
                    "id": template_id,
                    "overwrite_object": overwrites_object,
                    "overwrite_key": x.id,
                    "overwrite_pair": pairs[0],
                    "overwrite_second_pair": pairs[1],
                },
            )

    async def create_template(self, guild: discord.Guild) -> Template:
        self._check_database()

        template_id = str(uuid.uuid4())

        await self.database.insert(
            self.tables["templates"],
            {
                "guild": guild.id,
                "id": template_id,
                "afk_timeout": guild.afk_timeout,
                "mfa_level": guild.mfa_level,
                "verification_level": guild.verification_level.value,
                "explict_content_filter": guild.explicit_content_filter.value,
                "system_channel": guild.system_channel and guild.system_channel.id,
                "afk_channel": guild.afk_channel and guild.afk_channel.id,
            },
        )

        for category in guild.categories:
            await self.write_overwrites(template_id, category.id, category.overwrites)
            await self.database.insert(
                self.tables["categories"],
                {
                    "id": template_id,
                    "name": category.name,
                    "position": category.position,
                    "category_id": category.id,
                },
            )

        for channel in guild.text_channels:
            await self.write_overwrites(template_id, channel.id, channel.overwrites)
            await self.database.insert(
                self.tables["text_channels"],
                {
                    "id": template_id,
                    "name": channel.name,
                    "position": channel.position,
                    "category": channel.category_id,
                    "topic": channel.topic,
                    "slowmode": channel.slowmode_delay,
                    "nsfw": int(channel.is_nsfw()),
                    "channel_id": channel.id,
                },
            )

        for voice_channel in guild.voice_channels:
            await self.write_overwrites(
                template_id, voice_channel.id, voice_channel.overwrites
            )
            await self.database.insert(
                self.tables["voice_channels"],
                {
                    "id": template_id,
                    "name": voice_channel.name,
                    "position": voice_channel.position,
                    "category": voice_channel.category_id,
                    "bitrate": voice_channel.bitrate,
                    "user_limit": voice_channel.user_limit,
                    "channel_id": voice_channel.id,
                },
            )

        for role in guild.roles:
            await self.database.insert(
                self.tables["roles"],
                {
                    "id": template_id,
                    "default_role": int(role.is_default()),
                    "name": role.name,
                    "color": role.color.value,
                    "hoist": int(role.hoist),
                    "position": role.position,
                    "mentionable": int(role.mentionable),
                    "role_id": role.id,
                    "permissions": role.permissions.value,
                },
            )

        return await Template.get_template(self.database, self.tables, template_id)



from __future__ import annotations

import asyncio
import random
import re
import time
from enum import Enum
from typing import Optional, TYPE_CHECKING, List, Union, Any, Tuple

import aiohttp
import discord

from .exceptions import (
    QueueEmpty,
    NotPlaying,
    NotConnected,
    QueueError,
    AlreadyPaused,
    NotPaused,
    InvalidSkipIndex,
    SkipError,
    AlreadyConnected,
    UserNotConnected,
)
from .player import Player
from ..base import EventManager
from ..spotify import SpotifyClient

if TYPE_CHECKING:
    from discord.ext import commands


__all__ = (
    "Loops",
    "QueueManager",
    "MusicManager",
)

FFMPEG_OPTIONS = {
    "before_options": "-reconnect 1 -reconnect_streamed 1 -reconnect_delay_max 5",
    "options": "-vn",
}
SPOTIFY_RE = re.compile("^https://open.spotify.com/")


class Loops(Enum):
    NO_LOOP = 0
    LOOP = 1
    QUEUE_LOOP = 2


class QueueManager:
    __slots__ = (
        "queue",
        "volume",
        "history",
        "loop",
        "now_playing",
        "autoplay",
        "shuffle",
    )

    def __init__(self, volume: float, queue: List[Player]):
        self.queue = queue
        self.volume = volume
        self.history = []
        self.autoplay = False
        self.shuffle = False
        self.loop = Loops.NO_LOOP
        self.now_playing = None

    def add(self, player: Player) -> None:
        """
        Adds a player to the queue.

        :param player: The player to add.
        :type player: Player
        :return: None
        :rtype: None
        """

        self.queue.append(player)

    def clear(self) -> None:
        """
        Clears the queue.

        :return: None
        :rtype: None
        """

        self.queue.clear()

    def remove(self, index: int) -> Union[Player, Any]:
        """
        Removes and element from the queue at the specified index, and returns the element's value.

        :param index: The index.
        :type index: int
        :return: The element's value
        :rtype: Union[Player, Any]
        """

        return self.queue.pop(index)

    def cleanup(self):
        """
        Clears the queue.

        :return: None
        :rtype: None
        """

        self.clear()
        self.history.clear()
        del self.history
        del self.queue


class MusicManager(EventManager):
    """
    Represents a MusicManager.
    """

    __slots__ = (
        "bot",
        "client_id",
        "client_secret",
        "spotify_support",
        "inactivity_timeout",
        "queue",
        "spotify",
    )

    def __init__(
        self,
        bot: commands.Bot,
        spotify_support: bool = True,
        inactivity_timeout: int = 60,
        minimum_users: int = 1,
        **kwargs,
    ):
        super().__init__()
        self.bot = bot
        self.bot.add_listener(self.__on_voice_state_update, "on_voice_state_update")

        self.client_id = kwargs.get("client_id")
        self.client_secret = kwargs.get("client_secret")
        self.spotify_support = spotify_support
        self.inactivity_timeout = inactivity_timeout
        self.minimum_users = minimum_users

        self.queue = {}

        if spotify_support:
            self.spotify = SpotifyClient(
                client_id=self.client_id,
                client_secret=self.client_secret,
                loop=self.bot.loop,
            )

    async def cleanup(
        self, voice_client: Optional[discord.VoiceClient], guild: discord.Guild
    ) -> None:
        """
        |coro|

        Cleans up after a guild.

        :param Optional[discord.VoiceClient] voice_client: The voice client.
        :param discord.Guild guild: The guild.
        :return: None
        :rtype: None
        """

        if voice_client:
            await voice_client.disconnect()

        if guild.id in self.queue:
            self.queue[guild.id].cleanup()
            del self.queue[guild.id]

    async def __on_voice_state_update(self, member, before, after):
        voice_client = member.guild.voice_client
        channel_change = before.channel != after.channel

        if member == self.bot.user and channel_change:
            await self.cleanup(voice_client, member.guild)

        elif voice_client and channel_change:
            voice_members = list(
                filter(lambda x: not x.bot, voice_client.channel.members)
            )

            if len(voice_members) < self.minimum_users:
                await asyncio.sleep(self.inactivity_timeout)

                await self.cleanup(voice_client, member.guild)

    async def ensure_activity(self, ctx: commands.Context) -> None:
        """
        |coro|

        Waits the inactivity timeout and ensures the voice client in ctx is playing a song.
        If no song is playing, it disconnects and calls the on_inactivity_timeout event.

        :param ctx: The context.
        :type ctx: commands.Context
        :return: None
        :rtype: None
        """

        if self.inactivity_timeout is None:
            return

        await asyncio.sleep(self.inactivity_timeout)

        if not ctx.voice_client:
            return

        if ctx.voice_client.is_connected() and not ctx.voice_client.is_playing():
            await self.cleanup(ctx.voice_client, ctx.guild)

            await self.call_event("on_inactivity_disconnect", ctx)

    async def __check_connection(
        self,
        ctx: commands.Context,
        check_playing: bool = False,
        check_queue: bool = False,
    ) -> Optional[bool]:
        """
        |coro|

        Checks the connection state of the voice client in ctx.

        :param ctx: The context.
        :type ctx: commands.Context
        :param check_playing: A bool indicating if the function should check if a song is playing.
        :type check_playing: bool
        :param check_queue: A bool indicating if the function should check if a queue exists.
        :type check_queue: bool
        :return: True if all the checks passed.
        :rtype: bool
        """

        if not ctx.voice_client or not ctx.voice_client.is_connected():
            await self.call_event(
                "on_music_error",
                ctx,
                NotConnected("Client is not connected to a voice channel"),
            )
            return

        if check_playing and not ctx.voice_client.is_playing():
            await self.call_event(
                "on_music_error",
                ctx,
                NotPlaying("Client is not playing anything currently"),
            )
            return

        if check_queue and ctx.guild.id not in self.queue:
            await self.call_event("on_music_error", ctx, QueueEmpty("Queue is empty"))
            return

        return True

    @staticmethod
    async def _get_next_player(queue: QueueManager) -> Player:
        """
        |coro|

        Returns the next player that should be played from the queue.

        :param QueueManager queue: The queue.
        :return: The player.
        :rtype: Player
        """

        if queue.loop == Loops.LOOP:
            player = queue.now_playing

        elif queue.loop == Loops.QUEUE_LOOP:
            player = random.choice(queue.queue) if queue.shuffle else queue.remove(0)
            queue.add(player)

        else:
            if not queue.queue and queue.autoplay:
                last_video_id = queue.history[-1].data["videoDetails"]["videoId"]
                player = (await Player.get_similar_videos(last_video_id))[0]

            else:
                player = (
                    random.choice(queue.queue) if queue.shuffle else queue.remove(0)
                )

        return player

    async def __check_queue(self, ctx: commands.Context) -> None:
        """
        |coro|

        Plays the next song in the queue, handles looping, queue looping, autoplay, etc.

        :param ctx: The context of the voice client.
        :type ctx: commands.Context
        :return: None
        :rtype: None
        """

        try:
            if not ctx.voice_client or not ctx.voice_client.is_connected():
                return

            queue = self.queue[ctx.guild.id]
            player = await self._get_next_player(queue)

            if player is None:
                return

            queue.now_playing = player

            player.source = discord.PCMVolumeTransformer(
                discord.FFmpegPCMAudio(player.stream_url, **FFMPEG_OPTIONS),
                queue.volume,
            )

            ctx.voice_client.play(
                player.source,
                after=lambda x: self.bot.loop.create_task(self.__check_queue(ctx)),
            )
            player.start_timestamp = time.time()

            queue.history.append(player)
            await self.call_event("on_play", ctx, player)

        except (IndexError, KeyError):
            await self.cleanup(None, ctx.guild)
            await self.call_event("on_queue_end", ctx)

    async def get_player_played_duration(
        self, ctx: commands.Context, player: Player
    ) -> Optional[float]:
        """
        |coro|

        Returns the played duration of a player.

        :param ctx: The context.
        :type ctx: commands.Context
        :param player: The player.
        :type player: Player
        :return: The played duration of the player in seconds.
        :rtype: Optional[float]
        """

        if not await self.__check_connection(ctx):
            return

        start_timestamp = player.start_timestamp
        if ctx.voice_client.is_paused():
            start_timestamp = (
                player.start_timestamp + time.time() - player.last_pause_timestamp
            )

        time_played = time.time() - start_timestamp
        return min(
            time_played, time_played if player.duration == "LIVE" else player.duration
        )

    async def create_player(
        self, query: str, requester: discord.Member
    ) -> List[Player]:
        """
        |coro|

        Creates a list of players from the query.
        This function supports Spotify and all YTDL supported links.

        :param requester: The requester.
        :type requester: discord.Member
        :param query: The query.
        :type query: str
        :return: The list of players.
        :rtype: List[Player]
        """

        if SPOTIFY_RE.match(query) and self.spotify_support:
            return await Player.make_multiple_players(
                [song for song in await self.spotify.get_songs(query)], requester
            )

        return await Player.make_players(query, requester)

    async def queue_add(
        self, players: List[Player], ctx: commands.Context
    ) -> Optional[bool]:
        """
        |coro|

        Adds a list of players to the ctx queue.
        If a queue does not exist in ctx, it creates one.

        :param players: The list of players.
        :type players: List[Player]
        :param ctx: The context.
        :type ctx: commands.Context
        :return: A bool indicating if it was successful
        :rtype: Optional[bool]
        """

        if not await self.__check_connection(ctx):
            return

        if ctx.guild.id in self.queue:
            self.queue[ctx.guild.id].queue += players
        else:
            self.queue[ctx.guild.id] = QueueManager(0.1, players)

        return True

    async def queue_remove(self, ctx: commands.Context, index: int) -> None:
        """
        |coro|

        Removes a player from the queue in ctx at the specified index.
        Calls on_music_error with QueueError if index is invalid.

        :param ctx: The context.
        :type ctx: commands.Context
        :param index: The index.
        :type index: int
        :return: None
        :rtype: None
        """

        if not await self.__check_connection(ctx, check_queue=True):
            return

        try:
            self.queue[ctx.guild.id].remove(index)
        except IndexError:
            await self.call_event(
                "on_music_error",
                ctx,
                QueueError("Failure when removing player from queue"),
            )

    async def lyrics(
        self, ctx: commands.Context, query: str = None
    ) -> Optional[Tuple[str, str, str]]:
        """
        |coro|

        Returns the lyrics from the query or the currently playing song.

        :param ctx: The context.
        :type ctx: commands.Context
        :param query: The query.
        :type query: str
        :return: The lyrics and the song name.
        :rtype: Optional[Tuple[str, str, str]]
        """

        query = await self.now_playing(ctx) if query is None else query
        if not query:
            return

        url = f"https://some-random-api.ml/lyrics?title={query}"

        async with aiohttp.ClientSession() as session:
            request = await session.get(url)
            request_json = await request.json()

            authors = request_json.get("author")
            title = request_json.get("title")
            lyrics = request_json.get("lyrics")

            return (title, authors, lyrics) if lyrics else None

    async def play(
        self, ctx: commands.Context, player: Player = None
    ) -> Optional[bool]:
        """
        |coro|

        Plays the player or the next song in the queue if the player is not passed.

        :param ctx: The context.
        :type ctx: commands.Context
        :param player: The player.
        :type player: Player
        :return: A bool indicating if the play was successful
        :rtype: Optional[bool]
        """

        if not await self.__check_connection(ctx):
            return

        if player is not None:
            ctx.voice_client.play(player.source)
            return True

        if not ctx.voice_client.is_playing():
            await self.__check_queue(ctx)
            return True

    async def pause(self, ctx: commands.Context) -> Optional[bool]:
        """
        |coro|

        Pauses the currently playing song in ctx.
        Calls on_music_error with AlreadyPaused if already paused.

        :param ctx: The context.
        :type ctx: commands.Context
        :return: A bool indicating if the pause was successful
        :rtype: Optional[bool]
        """

        if not await self.__check_connection(ctx):
            return

        if ctx.voice_client.is_paused():
            await self.call_event(
                "on_music_error", ctx, AlreadyPaused("Player is already paused.")
            )
            return

        (await self.now_playing(ctx)).last_pause_timestamp = time.time()
        ctx.voice_client.pause()
        self.bot.loop.create_task(self.ensure_activity(ctx))
        return True

    async def resume(self, ctx: commands.Context) -> Optional[bool]:
        """
        |coro|

        Resumes the currently paused song in ctx.
        Calls on_music_error with NotPaused if not paused.

        :param ctx: The context.
        :type ctx: commands.Context
        :return: A bool indicating if the resume was successful
        :rtype: Optional[bool]
        """

        if not await self.__check_connection(ctx):
            return

        if not ctx.voice_client.is_paused():
            await self.call_event(
                "on_music_error", ctx, NotPaused("Player is not paused")
            )
            return

        ctx.voice_client.resume()

        now_playing = await self.now_playing(ctx)
        now_playing.start_timestamp += time.time() - now_playing.last_pause_timestamp

        return True

    async def skip(self, ctx: commands.Context, index: int = None) -> Optional[Player]:
        """
        |coro|

        Skips to the index in ctx.
        Calls on_music_error with InvalidSkipIndex or SkipError.

        :param index: The index to skip to.
        :type index: int
        :param ctx: The context.
        :type ctx: commands.Context
        :return: A bool indicating if the skip was successful
        :rtype: Optional[Player]
        """

        if not await self.__check_connection(ctx, True, check_queue=True):
            return

        queue = self.queue[ctx.guild.id]

        # Created duplicate to make sure InvalidSkipIndex isn't raised when the user does pass an index and the queue
        # is empty.
        skip_index = 0 if index is None else index - 1
        if not -1 < skip_index < len(queue.queue):
            if index:
                await self.call_event(
                    "on_music_error", ctx, InvalidSkipIndex("Skip index invalid.")
                )
                return

        if not queue.autoplay and len(queue.queue) <= skip_index:
            await self.call_event(
                "on_music_error", ctx, SkipError("No song to skip to.")
            )
            return

        if skip_index > 0:
            removed_songs = queue.queue[:skip_index]

            queue.queue = queue.queue[skip_index:]
            if queue.loop == Loops.QUEUE_LOOP:
                queue.queue += removed_songs

        if queue.autoplay:
            last_video_id = queue.history[-1].data["videoDetails"]["videoId"]
            player = (await Player.get_similar_videos(last_video_id))[0]
            queue.add(player)
        else:
            player = queue.queue[0]

        ctx.voice_client.stop()
        return player

    async def volume(
        self, ctx: commands.Context, volume: int = None
    ) -> Optional[float]:
        """
        |coro|

        Sets the volume in ctx.
        Returns the current volume if volume is None.

        :param volume: The volume to set.
        :type volume: int
        :param ctx: The context.
        :type ctx: commands.Context
        :return: The new volume.
        :rtype: Optional[float]
        """

        if not await self.__check_connection(ctx, True, check_queue=True):
            return

        if volume is None:
            return ctx.voice_client.source.volume * 100

        ctx.voice_client.source.volume = volume / 100
        self.queue[ctx.guild.id].volume = volume / 100
        return ctx.voice_client.source.volume * 100

    async def join(self, ctx: commands.Context) -> Optional[discord.VoiceChannel]:
        """
        |coro|

        Joins the ctx voice channel.
        Calls on_music_error with AlreadyConnected or UserNotConnected.

        :param ctx: The context.
        :type ctx: commands.Context
        :return: The voice channel it joined.
        :rtype: Optional[discord.VoiceChannel]
        """

        if ctx.voice_client and ctx.voice_client.is_connected():
            await self.call_event(
                "on_music_error",
                ctx,
                AlreadyConnected("Client is already connected to a voice channel"),
            )
            return

        if not ctx.author.voice:
            await self.call_event(
                "on_music_error",
                ctx,
                UserNotConnected("User is not connected to a voice channel"),
            )
            return

        channel = ctx.author.voice.channel
        await channel.connect()
        return channel

    async def leave(self, ctx: commands.Context) -> Optional[discord.VoiceChannel]:
        """
        |coro|

        Leaves the voice channel in ctx.

        :param ctx: The context.
        :type ctx: commands.Context
        :return: The voice channel it left.
        :rtype: Optional[discord.VoiceChannel]
        """

        if not await self.__check_connection(ctx):
            return

        if ctx.guild.id in self.queue:
            self.queue[ctx.guild.id].cleanup()
            del self.queue[ctx.guild.id]

        channel = ctx.voice_client.channel
        await ctx.voice_client.disconnect()
        return channel

    async def now_playing(self, ctx: commands.Context) -> Optional[Player]:
        """
        |coro|

        Returns the currently playing player.

        :param ctx: The context.
        :type ctx: commands.Context
        :return: The currently playing player.
        :rtype: Optional[Player]
        """

        if not await self.__check_connection(ctx, check_queue=True):
            return

        now_playing = self.queue[ctx.guild.id].now_playing
        if not ctx.voice_client.is_playing() and not ctx.voice_client.is_paused():
            await self.call_event(
                "on_music_error",
                ctx,
                NotPlaying("Client is not playing anything currently"),
            )

        return now_playing

    async def queueloop(self, ctx: commands.Context) -> Optional[bool]:
        """
        |coro|

        Toggles the queue loop.

        :param ctx: The context
        :type ctx: commands.Context
        :return: A bool indicating if the queue loop is now enabled or disabled.
        :rtype: Optional[bool]
        """

        if not await self.__check_connection(ctx, check_playing=True, check_queue=True):
            return

        self.queue[ctx.guild.id].loop = (
            Loops.QUEUE_LOOP
            if self.queue[ctx.guild.id].loop != Loops.QUEUE_LOOP
            else Loops.NO_LOOP
        )

        if self.queue[ctx.guild.id].loop == Loops.QUEUE_LOOP:
            self.queue[ctx.guild.id].add(self.queue[ctx.guild.id].now_playing)

        return self.queue[ctx.guild.id].loop == Loops.QUEUE_LOOP

    async def shuffle(self, ctx: commands.Context) -> Optional[bool]:
        """
        |coro|

        Toggles the shuffle feature.

        :param commands.Context ctx: The context
        :return: A bool indicating if the queue loop is now enabled or disabled.
        :rtype: Optional[bool]
        """

        if not await self.__check_connection(ctx, check_playing=True, check_queue=True):
            return

        self.queue[ctx.guild.id].shuffle = not self.queue[ctx.guild.id].shuffle
        return self.queue[ctx.guild.id].shuffle

    async def autoplay(self, ctx: commands.Context) -> Optional[bool]:
        """
        |coro|

        Toggles the autoplay feature.

        :param commands.Context ctx: The context
        :return: A bool indicating if autoplay is now enabled or disabled.
        :rtype: Optional[bool]
        """

        if not await self.__check_connection(ctx, check_playing=True, check_queue=True):
            return

        self.queue[ctx.guild.id].autoplay = not self.queue[ctx.guild.id].autoplay
        return self.queue[ctx.guild.id].autoplay

    async def loop(self, ctx: commands.Context) -> Optional[bool]:
        """
        |coro|

        Toggles the loop.

        :param ctx: The context
        :type ctx: commands.Context
        :return: A bool indicating if the loop is now enabled or disabled.
        :rtype: Optional[bool]
        """

        if not await self.__check_connection(ctx, check_playing=True, check_queue=True):
            return

        self.queue[ctx.guild.id].loop = (
            Loops.LOOP if self.queue[ctx.guild.id].loop != Loops.LOOP else Loops.NO_LOOP
        )
        return self.queue[ctx.guild.id].loop == Loops.LOOP

    async def get_queue(self, ctx: commands.Context) -> Optional[QueueManager]:
        """
        |coro|

        Returns the queue of ctx.

        :param ctx: The context.
        :type ctx: commands.Context
        :return: The queue.
        :rtype: Optional[QueueManager]
        """

        if not await self.__check_connection(ctx, check_queue=True):
            return

        return self.queue[ctx.guild.id]



import asyncio
from math import ceil

import discord

__all__ = ("generate_embeds", "EmojiError", "PageManager", "ButtonsPageManager")


def generate_embeds(
    list_to_generate, title, description, fields=25, color=0xFF0000, string_format="{}"
):
    num_of_embeds = ceil((len(list_to_generate) + 1) / fields)

    embeds = [
        discord.Embed(
            title=f"{title} (Page 1/{num_of_embeds})",
            description=description,
            color=color,
        )
    ]

    for i in range(2, num_of_embeds + 1):
        embeds.append(
            discord.Embed(title=f"{title} (Page {i}/{num_of_embeds})", color=color)
        )

    embed_index = 0
    for index, element in enumerate(list_to_generate):
        embeds[embed_index].add_field(
            name=f"**{index + 1}.**", value=string_format.format(element), inline=False
        )

        if (index + 1) % fields == 0:
            embed_index += 1

    return embeds


class ButtonError(Exception):
    __slots__ = ()


class EmojiError(Exception):
    __slots__ = ()


class ButtonsPageManager:
    __slots__ = (
        "ctx",
        "messages",
        "timeout",
        "buttons",
        "public",
        "index",
        "button_color",
    )

    def __init__(
        self,
        ctx,
        messages,
        timeout=60,
        buttons=None,
        public=False,
        index=0,
        button_color=None,
    ):
        self.ctx = ctx
        self.messages = messages
        self.timeout = timeout
        self.buttons = buttons if buttons is not None else ["⏪", "◀️", "▶️", "⏩"]
        self.public = public
        self.index = index
        self.button_color = button_color

    async def run(self):
        if len(self.buttons) != 4:
            raise ButtonError(f"Passed {len(self.buttons)} buttons when 4 are needed.")

        self.index = 0 if not -1 < self.index < len(self.messages) else self.index

        from discord_components import ActionRow, Button, ButtonStyle, DiscordComponents

        DiscordComponents(self.ctx.bot)

        components = ActionRow(
            [
                Button(
                    style=self.button_color or ButtonStyle.red,
                    label=button,
                    custom_id=button,
                )
                for button in self.buttons
            ]
        )

        message_to_send = self.messages[self.index]
        # message_to_send must be of type embed, sadly, discord_components breaks the Messageable.send method
        # And breaks the file parameter, too
        message = await self.ctx.send(embed=message_to_send, components=components)

        while True:
            try:
                interaction = await self.ctx.bot.wait_for(
                    "button_click", check=lambda i: i.message == message, timeout=30
                )

                if interaction.user.bot:
                    continue

                if interaction.user != self.ctx.author and not self.public:
                    continue

            except asyncio.TimeoutError:
                break

            if interaction.custom_id == self.buttons[0]:
                self.index = 0

            elif interaction.custom_id == self.buttons[1]:
                if self.index > 0:
                    self.index -= 1

            elif interaction.custom_id == self.buttons[2]:
                if self.index < len(self.messages) - 1:
                    self.index += 1

            elif interaction.custom_id == self.buttons[3]:
                self.index = len(self.messages) - 1

            message_to_send = self.messages[self.index]

            for button in components[0]:
                button.disabled = False

            if self.index == len(self.messages) - 1:
                components[0][2].disabled = True
                components[0][3].disabled = True

            if self.index == 0:
                components[0][0].disabled = True
                components[0][1].disabled = True

            await interaction.respond(
                type=7,  # Cannot find a proper enum :)
                content="",
                embed=message_to_send,
                components=components,
            )


class PageManager:
    __slots__ = ("ctx", "messages", "timeout", "emojis", "public", "index")

    def __init__(self, ctx, messages, timeout=60, emojis=None, public=False, index=0):
        self.ctx = ctx
        self.messages = messages
        self.timeout = timeout
        self.emojis = emojis if emojis is not None else ["⏪", "◀️", "▶️", "⏩"]
        self.public = public
        self.index = index

    async def run(self):
        if len(self.emojis) != 4:
            raise EmojiError(f"Passed {len(self.emojis)} emojis when 4 are needed.")

        self.index = 0 if not -1 < self.index < len(self.messages) else self.index

        message_to_send = self.messages[self.index]
        if isinstance(message_to_send, discord.Embed):
            message = await self.ctx.send(embed=message_to_send)
        else:
            message = await self.ctx.send(message_to_send)

        for emoji in self.emojis:
            await message.add_reaction(emoji)

        while True:
            try:
                reaction, user = await self.ctx.bot.wait_for(
                    "reaction_add",
                    check=lambda x, y: x.message == message,
                    timeout=self.timeout,
                )

                if user.bot:
                    continue

                if user != self.ctx.author and not self.public:
                    continue

            except asyncio.TimeoutError:
                break

            if reaction.emoji == self.emojis[0]:
                self.index = 0

            elif reaction.emoji == self.emojis[1]:
                if self.index > 0:
                    self.index -= 1

            elif reaction.emoji == self.emojis[2]:
                if self.index < len(self.messages) - 1:
                    self.index += 1

            elif reaction.emoji == self.emojis[3]:
                self.index = len(self.messages) - 1

            await message.remove_reaction(reaction.emoji, user)
            message_to_send = self.messages[self.index]

            if isinstance(message_to_send, discord.Embed):
                await message.edit(embed=message_to_send, content=None)
            else:
                await message.edit(content=message_to_send, embed=None)



from __future__ import annotations

import asyncio
from datetime import datetime, timedelta
from typing import Dict, List, Optional, Any

import discord
import pytz
from discord.ext import commands

from .base import DatabaseChecker


class PartialBirthdayMember:
    def __init__(self, member: discord.Member, birthday_date: datetime, timezone: str):
        self.member = member
        self.birthday_date = birthday_date
        self.timezone = timezone


class BirthdayMember:
    def __init__(self, birthday_manager: BirthdayManager, member: discord.Member):
        self.birthday_manager = birthday_manager
        self.member = member
        self.table = self.birthday_manager.tables["birthdays"]

    @property
    def __checks(self) -> Dict[str, int]:
        return {"guild": self.member.guild.id, "member": self.member.id}

    async def birthday_date(self) -> datetime:
        birthday_data = await self.birthday_manager.database.select(
            self.table, ["utc_birthday"], self.__checks
        )
        return datetime.utcfromtimestamp(birthday_data["utc_birthday"])

    async def next_birthday(self) -> datetime:
        current_datetime = datetime.utcnow()

        new_date = (await self.birthday_date()).replace(year=current_datetime.year)
        if new_date.timestamp() - current_datetime.timestamp() < 0:
            new_date = new_date.replace(year=current_datetime.year + 1)

        return new_date

    async def timezone(self) -> str:
        timezone_data = await self.birthday_manager.database.select(
            self.table, ["timezone"], self.__checks
        )
        return timezone_data["timezone"]

    async def delete(self) -> PartialBirthdayMember:
        partial = PartialBirthdayMember(
            self.member, await self.birthday_date(), await self.timezone()
        )
        await self.birthday_manager.database.delete(self.table, self.__checks)
        return partial

    async def set_birthday_date(self, timestamp: float) -> None:
        await self.birthday_manager.database.update(
            self.table, {"utc_birthday": timestamp}, self.__checks
        )

    async def set_timezone(self, timezone: str) -> None:
        await self.birthday_manager.database.update(
            self.table, {"timezone": timezone}, self.__checks
        )

    async def age(self) -> int:
        current_birthday = await self.birthday_date()
        current_datetime = datetime.utcnow()

        return (
            current_datetime.year
            - current_birthday.year
            - (
                datetime(
                    year=current_datetime.year,
                    month=current_birthday.month,
                    day=current_birthday.day,
                )
                >= current_datetime
            )
        )


class BirthdayManager(DatabaseChecker):
    def __init__(self, bot: commands.Bot):
        super().__init__(
            [
                {
                    "guild": "snowflake",
                    "member": "snowflake",
                    "utc_birthday": "snowflake",
                    "timezone": "string",
                }
            ],
            ["birthdays"],
        )
        self.bot = bot
        self.add_event(self.on_database_connect)

    async def on_database_connect(self):
        self.bot.loop.create_task(self.__detect_birthdays())

    async def create_birthday(
        self, member: discord.Member, member_birthday: float, timezone: str = "UTC"
    ) -> None:
        self._check_database()

        await self.database.insertifnotexists(
            self.tables["birthdays"],
            dict(
                zip(
                    self.tables_column_data[0],
                    [member.guild.id, member.id, member_birthday, timezone],
                )
            ),
            {"guild": member.guild.id, "member": member.id},
        )

    async def get_birthday(self, member: discord.Member) -> Optional[BirthdayMember]:
        self._check_database()

        member_data = await self.database.select(
            self.tables["birthdays"],
            [],
            {"guild": member.guild.id, "member": member.id},
            True,
        )

        if member_data:
            return BirthdayMember(self, member)

        return None

    async def get_upcoming(self, guild: discord.Guild) -> List[BirthdayMember]:
        self._check_database()

        member_data = await self.database.select(
            self.tables["birthdays"], [], fetchall=True
        )

        birthdays = {}

        for birthday_member in member_data:
            member = guild.get_member(birthday_member["member"])

            if member:
                birthday = BirthdayMember(self, member)
                birthdays[birthday] = await birthday.next_birthday()
                # Did this instead of running the method until complete in the sorted key lambda

        return sorted(birthdays, key=lambda x: birthdays[x])

    @staticmethod
    def get_midnight_timezones() -> List[str]:
        """
        This method returns a list of timezones where the current time is 12 am.
        :return:
        """

        current_utc_time = datetime.utcnow()
        utc_offset = -(current_utc_time.hour % 24)

        minutes = 30 if current_utc_time.minute > 5 else 0
        if minutes == 30:
            utc_offset -= 1

        checks = (
            timedelta(hours=utc_offset, minutes=minutes),
            timedelta(
                hours=24 - -utc_offset if current_utc_time.hour != 0 else 0,
                minutes=minutes,
            ),
        )

        return [
            tz.zone
            for tz in map(pytz.timezone, pytz.all_timezones_set)
            if current_utc_time.astimezone(tz).utcoffset() in checks
        ]

    async def get_members_with_birthday(
        self, timezones: List[str]
    ) -> List[Dict[str, Any]]:
        """
        This function receives a list of timezones and returns a list of members that have birthdays in that date
        and timezone.

        :param timezones:
        :return:
        """

        result_members = []
        registered_members = await self.database.select(
            self.tables["birthdays"], [], fetchall=True
        )

        birthday_members = [x for x in registered_members if x["timezone"] in timezones]
        for birthday_member in birthday_members:
            timezone_time = datetime.now(pytz.timezone(birthday_member["timezone"]))
            date_of_birth = datetime.fromtimestamp(birthday_member["utc_birthday"])

            if (
                date_of_birth.month == timezone_time.month
                and date_of_birth.day == timezone_time.day
            ):
                result_members.append(birthday_member)

        return result_members

    @staticmethod
    def round_to_nearest(timedelta_to_round):
        """
        This function receives a timedelta to round to and gets the amount of seconds before that timestamp.

        :param timedelta_to_round:
        :return:
        """

        now = datetime.now()
        nearest = now + (datetime.min - now) % timedelta_to_round
        return nearest.timestamp() - now.timestamp()

    async def __detect_birthdays(self) -> None:
        await self.bot.wait_until_ready()

        while not self.bot.is_closed():
            await asyncio.sleep(self.round_to_nearest(timedelta(minutes=30)))

            for birthday_member in await self.get_members_with_birthday(
                self.get_midnight_timezones()
            ):
                guild = self.bot.get_guild(birthday_member["guild"])

                if guild:
                    member = guild.get_member(birthday_member["member"])

                    if member:
                        await self.call_event(
                            "on_member_birthday", BirthdayMember(self, member)
                        )



""""
If InviteTracker is used in any way that breaks Discord TOS we, (the DiscordSuperUtils team)
are not responsible or liable in any way.
InviteTracker by DiscordSuperUtils was not intended to violate Discord TOS in any way.
In case we are contacted by Discord, we will remove any and all features that violate the Discord ToS.
Please feel free to read the Discord Terms of Service https://discord.com/terms.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Union, Optional

from .base import DatabaseChecker

if TYPE_CHECKING:
    import discord
    from discord.ext import commands


class InviteAccount:
    def __init__(self, invite_tracker: InviteTracker, member: discord.Member):
        self.invite_tracker = invite_tracker
        self.member = member

    def __str__(self):
        return f"<member={self.member.id}>"

    async def get_invited_users(self):
        return await self.invite_tracker.get_members_invited(
            self.member, self.member.guild
        )


class InviteTracker(DatabaseChecker):
    def __init__(self, bot: commands.Bot):
        super().__init__(
            [
                {
                    "guild": "snowflake",
                    "member": "snowflake",
                    "members_invited": "string",
                }
            ],
            ["invites"],
        )
        self.bot = bot
        self.cache = {}

        self.bot.loop.create_task(self.__initialize_cache())

        self.bot.add_listener(self.__cleanup_guild_cache, "on_guild_remove")
        self.bot.add_listener(self.__update_guild_cache, "on_guild_add")
        self.bot.add_listener(self.__track_invite, "on_invite_create")
        self.bot.add_listener(self.__cleanup_invite, "on_invite_delete")

    async def get_invite(self, member: discord.Member) -> Optional[discord.Invite]:
        for inv in await member.guild.invites():
            for invite in self.cache[member.guild.id]:
                if invite.revoked:
                    self.cache[invite.guild.id].remove(invite)
                    return

                if invite.code == inv.code and inv.uses - invite.uses == 1:
                    await self.__update_guild_cache(member.guild)
                    return inv

    async def get_members_invited(
        self, user: Union[discord.User, discord.Member], guild: discord.Guild
    ):
        self._check_database()

        invited_members = await self.database.select(
            self.tables["invites"],
            ["members_invited"],
            {"guild": guild.id, "member": user.id},
        )

        if not invited_members:
            return []

        invited_members = invited_members["members_invited"]
        if isinstance(invited_members, str):
            return [
                int(invited_member)
                for invited_member in invited_members.split("\0")
                if invited_member
            ]

    async def fetch_inviter(
        self, invite: discord.Invite
    ) -> Union[discord.Member, discord.User]:
        inviter = invite.guild.get_member(invite.inviter.id)
        return inviter if inviter else await self.bot.get_user(invite.inviter.id)

    async def register_invite(
        self,
        invite: discord.Invite,
        member: discord.Member,
        inviter: Union[discord.Member, discord.User],
    ) -> None:
        self._check_database()

        invited_members = await self.get_members_invited(inviter, invite.guild)
        if member.id in invited_members:
            return

        invited_members.append(member.id)
        invited_members_sql = "\0".join(
            str(invited_member) for invited_member in invited_members
        )

        await self.database.updateorinsert(
            self.tables["invites"],
            {"members_invited": invited_members_sql},
            {"guild": invite.guild.id, "member": inviter.id},
            {
                "guild": invite.guild.id,
                "member": inviter.id,
                "members_invited": invited_members_sql,
            },
        )

    async def __initialize_cache(self) -> None:
        await self.bot.wait_until_ready()

        for guild in self.bot.guilds:
            self.cache[guild.id] = await guild.invites()

    async def __update_guild_cache(self, guild: discord.Guild) -> None:
        self.cache[guild.id] = await guild.invites()

    async def __track_invite(self, invite: discord.Invite) -> None:
        self.cache[invite.guild.id].append(invite)

    async def __cleanup_invite(self, invite: discord.Invite) -> None:
        if invite in self.cache[invite.guild.id]:
            self.cache[invite.guild.id].remove(invite)

    async def __cleanup_guild_cache(self, guild: discord.Guild) -> None:
        self.cache.pop(guild.id)

    def get_user_info(self, member: discord.Member) -> InviteAccount:
        self._check_database()

        return InviteAccount(self, member)



from abc import ABC, abstractmethod
from difflib import SequenceMatcher
from typing import List, Union

import discord
from discord.ext import commands

from .base import get_generator_response

__all__ = ("CommandResponseGenerator", "DefaultResponseGenerator", "CommandHinter")


class CommandResponseGenerator(ABC):
    __slots__ = ()

    @abstractmethod
    def generate(
        self, invalid_command: str, suggestions: List[str]
    ) -> Union[str, discord.Embed]:
        pass


class DefaultResponseGenerator(CommandResponseGenerator):
    __slots__ = ()

    def generate(self, invalid_command: str, suggestions: List[str]) -> discord.Embed:
        embed = discord.Embed(
            title="Invalid command!",
            description=f"**`{invalid_command}`** is invalid. Did you mean:",
            color=0x00FF00,
        )

        for index, suggestion in enumerate(suggestions[:3]):
            embed.add_field(
                name=f"**{index + 1}.**", value=f"**`{suggestion}`**", inline=False
            )

        return embed


class CommandHinter:
    __slots__ = ("bot", "generator")

    def __init__(self, bot: commands.Bot, generator=None):
        self.bot = bot
        self.generator = DefaultResponseGenerator if generator is None else generator

        self.bot.add_listener(self.__handle_hinter, "on_command_error")

    @property
    def command_names(self) -> List[str]:
        names = []

        for command in self.bot.commands:
            if isinstance(command, commands.Group):
                names += [command.name] + command.aliases
                for inner_command in command.commands:
                    names += [inner_command.name] + inner_command.aliases

            else:
                names += [command.name] + command.aliases

        return names

    async def __handle_hinter(self, ctx: commands.Context, error) -> None:
        if isinstance(error, commands.CommandNotFound):
            command_similarity = {}
            command_used = ctx.message.content.lstrip(ctx.prefix)[
                : max([len(c) for c in self.command_names])
            ]

            for command in self.command_names:
                command_similarity[
                    SequenceMatcher(None, command, command_used).ratio()
                ] = command

            generated_message = get_generator_response(
                self.generator,
                CommandResponseGenerator,
                command_used,
                [x[1] for x in sorted(command_similarity.items(), reverse=True)],
            )

            if not generated_message:
                return

            if isinstance(generated_message, discord.Embed):
                await ctx.send(embed=generated_message)
            elif isinstance(generated_message, str):
                await ctx.send(generated_message)
            else:
                raise TypeError(
                    "The generated message must be of type 'discord.Embed' or 'str'."
                )

        else:
            raise error



from __future__ import annotations

import os
import textwrap
from enum import Enum
from io import BytesIO
from typing import Optional, Tuple, Union, TYPE_CHECKING

import PIL
import PIL.ImageShow
import aiohttp
import discord
from PIL import Image, ImageFont, ImageDraw

if TYPE_CHECKING:
    from .leveling import LevelingAccount


__all__ = ("ImageManager", "Backgrounds")


class ImageManager:
    __slots__ = ()

    @staticmethod
    def load_asset(name: str) -> str:
        return os.path.join(os.path.dirname(__file__), "assets", name)

    @staticmethod
    async def make_request(url: str) -> Optional[bytes]:
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                return await response.read()

    @classmethod
    async def convert_image(cls, url: str) -> Image:
        return PIL.Image.open(BytesIO(await cls.make_request(url))).convert("RGBA")

    @classmethod
    def human_format(cls, num):
        original_num = num

        num = float("{:.3g}".format(num))
        magnitude = 0
        matches = ["", "K", "M", "B", "T", "Qua", "Qui"]
        while abs(num) >= 1000:
            if magnitude >= 5:
                break

            magnitude += 1
            num /= 1000.0

        try:
            return "{}{}".format(
                "{:f}".format(num).rstrip("0").rstrip("."), matches[magnitude]
            )
        except IndexError:
            return original_num

    @staticmethod
    def multiline_text(
        card: ImageDraw,
        text: str,
        font: ImageFont,
        text_color: Tuple[int, int, int],
        start_height: Union[int, float],
        width: int,
    ):
        draw = ImageDraw.Draw(card)
        image_width, image_height = card.size

        y_text = start_height
        lines = textwrap.wrap(text, width=width)

        for line in lines:
            line_width, line_height = font.getsize(line)
            draw.text(
                ((image_width - line_width) / 2, y_text),
                line,
                font=font,
                fill=text_color,
            )
            y_text += line_height

    async def draw_profile_picture(
        self,
        card: Image,
        member: discord.Member,
        location: Tuple[int, int],
        size: int = 180,
        outline_thickness: int = 5,
        status: bool = True,
        outline_color: Tuple[int, int, int] = (255, 255, 255),
    ):
        blank = Image.new("RGBA", card.size, (255, 255, 255, 0))

        location = tuple(
            round(x - size / 2) if i <= 1 else round(x + size / 2)
            for i, x in enumerate(location + location)
        )

        outline_dimensions = tuple(
            x - outline_thickness if i <= 1 else x + outline_thickness
            for i, x in enumerate(location)
        )

        size_dimensions = (size, size)
        status_dimensions = tuple(round(x / 4) for x in size_dimensions)

        mask = Image.new("RGBA", card.size, 0)
        ImageDraw.Draw(mask).ellipse(location, fill=(255, 25, 255, 255))

        avatar = (await self.convert_image(str(member.avatar_url))).resize(
            size_dimensions
        )
        profile_pic_holder = Image.new("RGBA", card.size, (255, 255, 255, 255))

        ImageDraw.Draw(card).ellipse(outline_dimensions, fill=outline_color)

        profile_pic_holder.paste(avatar, location)
        pre_card = Image.composite(profile_pic_holder, card, mask)
        pre_card = pre_card.convert("RGBA")

        if status:
            status_picture = Image.open(self.load_asset(f"{member.status.name}.png"))
            status_picture = status_picture.convert("RGBA").resize(status_dimensions)

            blank.paste(
                status_picture, tuple(x - status_dimensions[0] for x in location[2:])
            )

        return Image.alpha_composite(pre_card, blank)

    async def create_welcome_card(
        self,
        member: discord.Member,
        background: Union[Backgrounds, str],
        text_color: Tuple[int, int, int],
        title: str,
        description: str,
        font_path: str = None,
        outline: int = 5,
        transparency: Optional[int] = 0,
    ) -> discord.File:
        result_bytes = BytesIO()

        card = (
            Image.open(background.value)
            if isinstance(background, Backgrounds)
            else await self.convert_image(background)
        )
        card = card.resize((1024, 500))

        font_path = font_path if font_path else self.load_asset("font.ttf")

        big_font = ImageFont.truetype(font_path, 36)
        small_font = ImageFont.truetype(font_path, 30)

        draw = ImageDraw.Draw(card, "RGBA")
        if transparency:
            draw.rectangle((30, 30, 994, 470), fill=(0, 0, 0, transparency))

        draw.text((512, 360), title, text_color, font=big_font, anchor="ms")
        self.multiline_text(card, description, small_font, text_color, 380, 60)

        final_card = await self.draw_profile_picture(
            card, member, (512, 180), 260, outline_thickness=outline
        )

        final_card.save(result_bytes, format="PNG")
        result_bytes.seek(0)
        return discord.File(result_bytes, filename="welcome_card.png")

    async def merge_image(
        self,
        foreground: str,
        background: str,
        blend_level: float = 0.6,
        discord_file: bool = True,
    ) -> Union[discord.File, Image]:
        """Merges two images together"""
        foreground = await self.convert_image(foreground)
        background = await self.convert_image(background)
        result_bytes = BytesIO()
        width, height = background.size

        foreground = foreground.resize((width, height), PIL.Image.ANTIALIAS)
        result = PIL.Image.blend(background, foreground, alpha=blend_level)

        if discord_file:
            result.save(result_bytes, format="PNG")
            result_bytes.seek(0)
            return discord.File(result_bytes, filename="mergedimage.png")

        return result

    async def create_leveling_profile(
        self,
        member: discord.Member,
        member_account: LevelingAccount,
        background: Union[Backgrounds, str],
        text_color: Tuple[int, int, int],
        rank: int,
        font_path: str = None,
        outline: int = 5,
    ) -> discord.File:
        result_bytes = BytesIO()

        card = (
            Image.open(background.value)
            if isinstance(background, Backgrounds)
            else await self.convert_image(background)
        )
        card = card.resize((850, 238))

        font_path = font_path if font_path else self.load_asset("font.ttf")
        font_big = ImageFont.truetype(font_path, 36)
        font_small = ImageFont.truetype(font_path, 20)

        draw = ImageDraw.Draw(card)
        draw.text((245, 90), str(member), text_color, font=font_big, anchor="ls")
        draw.text((800, 90), f"Rank #{rank}", text_color, font=font_big, anchor="rs")
        draw.text(
            (245, 165),
            f"Level {await member_account.level()}",
            text_color,
            font=font_small,
            anchor="ls",
        )
        draw.text(
            (800, 165),
            f"{self.human_format(await member_account.xp())} /"
            f" {self.human_format(await member_account.next_level())} XP",
            text_color,
            font=font_small,
            anchor="rs",
        )

        draw.rounded_rectangle(
            (245, 185, 800, 205), fill=(0, 0, 0, 0), outline=text_color, radius=10
        )
        length_of_bar = await member_account.percentage_next_level() * 5.5 + 250
        draw.rounded_rectangle(
            (245, 185, length_of_bar, 205), fill=text_color, radius=10
        )

        final_card = await self.draw_profile_picture(
            card,
            member,
            (109, 119),
            outline_thickness=outline,
            outline_color=text_color,
        )

        final_card.save(result_bytes, format="PNG")
        result_bytes.seek(0)
        return discord.File(result_bytes, filename="rankcard.png")


class Backgrounds(Enum):
    GALAXY = ImageManager.load_asset("1.png")
    BLANK_GRAY = ImageManager.load_asset("2.png")
    GAMING = ImageManager.load_asset("3.png")



from __future__ import annotations

import asyncio
from datetime import datetime
from typing import TYPE_CHECKING, Union, Optional, List, Any, Dict

import discord
import discord.utils

from .base import DatabaseChecker
from .punishments import Punisher

if TYPE_CHECKING:
    from discord.ext import commands
    from .punishments import Punishment


__all__ = ("AlreadyMuted", "MuteManager")


class AlreadyMuted(Exception):
    """Raises an error when a user is already muted."""


class MuteManager(DatabaseChecker, Punisher):
    """
    A MuteManager that handles mutes for guilds.
    """

    __slots__ = ("bot",)

    def __init__(self, bot: commands.Bot):
        super().__init__(
            [
                {
                    "guild": "snowflake",
                    "member": "snowflake",
                    "timestamp_of_mute": "snowflake",
                    "timestamp_of_unmute": "snowflake",
                    "reason": "string",
                }
            ],
            ["mutes"],
        )
        self.bot = bot

        self.add_event(self.on_database_connect)

    async def on_database_connect(self):
        self.bot.loop.create_task(self.__check_mutes())
        self.bot.add_listener(self.on_member_join)

    async def get_muted_members(self) -> List[Dict[str, Any]]:
        """
        |coro|

        This function returns all the members that are supposed to be unmuted but are muted.

        :return: The unmuted members.
        :rtype: List[Dict[str, Any]]
        """

        return [
            x
            for x in await self.database.select(self.tables["mutes"], [], fetchall=True)
            if x["timestamp_of_unmute"] <= datetime.utcnow().timestamp()
        ]

    async def on_member_join(self, member: discord.Member) -> None:
        """
        |coro|

        The on_member_join event callback.
        Used so the member cant leave the guild, join back and be unmuted.

        :param member: The member that joined.
        :type member: discord.Member
        :return: None
        :rtype: None
        """

        muted_members = [
            x
            for x in await self.database.select(
                self.tables["mutes"],
                ["timestamp_of_unmute", "member"],
                {"guild": member.guild.id, "member": member.id},
                fetchall=True,
            )
            if x["timestamp_of_unmute"] > datetime.utcnow().timestamp()
        ]

        if any([muted_member["member"] == member.id for muted_member in muted_members]):
            muted_role = discord.utils.get(member.guild.roles, name="Muted")

            if muted_role:
                await member.add_roles(muted_role)

    async def __check_mutes(self) -> None:
        """
        |coro|

        A loop that makes sure the members are unmuted when they are supposed to.

        :return: None
        :rtype: None
        """

        await self.bot.wait_until_ready()

        while not self.bot.is_closed():
            for muted_member in await self.get_muted_members():
                guild = self.bot.get_guild(muted_member["guild"])

                if guild is None:
                    continue

                member = guild.get_member(muted_member["member"])

                if await self.unmute(member):
                    await self.call_event("on_unmute", member, muted_member["reason"])

            await asyncio.sleep(300)

    async def punish(
        self, ctx: commands.Context, member: discord.Member, punishment: Punishment
    ) -> None:
        try:
            await self.mute(member)
        except discord.errors.Forbidden as e:
            raise e
        else:
            await self.call_event("on_punishment", ctx, member, punishment)

    @staticmethod
    async def ensure_permissions(
        guild: discord.Guild, muted_role: discord.Role
    ) -> None:
        """
        |coro|

        This function loops through the guild's channels and ensures the muted_role is not allowed to
        send messages or speak in that channel.

        :param guild: The guild to get the channels from.
        :type guild: discord.Guild
        :param muted_role: The muted role.
        :type muted_role: discord.Role
        :return: None
        """

        channels_to_mute = [
            channel
            for channel in guild.channels
            if channel.overwrites_for(muted_role).send_messages is not False
        ]
        # Now, you might say what the heck, why don't you test if the value is True instead of checking if it
        # is not False? I am doing it this way because permissions have 3 values,
        # None, True and False.
        # Now, lets say we have a permission that is set to None, if i test it for a False value, (if not value) it will
        # return False which is incorrect and it should return True.

        await asyncio.gather(
            *[
                channel.set_permissions(muted_role, send_messages=False, speak=False)
                for channel in channels_to_mute
            ]
        )

    async def __handle_unmute(
        self, time_of_mute: Union[int, float], member: discord.Member, reason: str
    ) -> None:
        """
        |coro|

        A function that handles the member's unmute that runs separately from mute so it wont be blocked.

        :param time_of_mute: The time until the member's unmute timestamp.
        :type time_of_mute: Union[int, float]
        :param member: The member to unmute.
        :type member: discord.Member
        :param reason: The reason of the mute.
        :type reason: str
        :return: None
        """

        await asyncio.sleep(time_of_mute)

        if await self.unmute(member):
            await self.call_event("on_unmute", member, reason)

    async def mute(
        self,
        member: discord.Member,
        reason: str = "No reason provided.",
        time_of_mute: Union[int, float] = 0,
    ) -> None:
        """
        |coro|

        Mutes a member.

        :raises: AlreadyMuted: The member is already muted.
        :param member: The member to mute.
        :type member: discord.Member
        :param reason: The reason of the mute.
        :type reason: str
        :param time_of_mute: The time of mute.
        :type time_of_mute: Union[int, float]
        :return: None,
        :rtype: None
        """

        self._check_database()

        muted_role = discord.utils.get(member.guild.roles, name="Muted")
        if not muted_role:
            muted_role = await member.guild.create_role(
                name="Muted",
                permissions=discord.Permissions(send_messages=False, speak=False),
            )

        if muted_role in member.roles:
            raise AlreadyMuted(f"{member} is already muted.")

        await member.add_roles(muted_role, reason=reason)

        self.bot.loop.create_task(self.ensure_permissions(member.guild, muted_role))

        if time_of_mute <= 0:
            return

        await self.database.insert(
            self.tables["mutes"],
            {
                "guild": member.guild.id,
                "member": member.id,
                "timestamp_of_mute": datetime.utcnow().timestamp(),
                "timestamp_of_unmute": datetime.utcnow().timestamp() + time_of_mute,
                "reason": reason,
            },
        )

        self.bot.loop.create_task(self.__handle_unmute(time_of_mute, member, reason))

    async def unmute(self, member: discord.Member) -> Optional[bool]:
        """
        |coro|

        Unmutes a member.

        :param member: The member to unmute.
        :type member: discord.Member
        :rtype: Optional[bool]
        :return: A bool indicating if the unmute was successful
        """

        await self.database.delete(
            self.tables["mutes"], {"guild": member.guild.id, "member": member.id}
        )
        muted_role = discord.utils.get(member.guild.roles, name="Muted")
        if not muted_role:
            return

        if muted_role not in member.roles:
            return

        await member.remove_roles(muted_role)
        return True



from __future__ import annotations

import asyncio
from datetime import datetime
from typing import TYPE_CHECKING, Union, Optional, List, Dict, Any

import discord

from .base import DatabaseChecker
from .punishments import Punisher

if TYPE_CHECKING:
    from .punishments import Punishment
    from discord.ext import commands


__all__ = ("UnbanFailure", "BanManager")


class UnbanFailure(Exception):
    """Raises an exception when the user tries to unban a discord.User without passing the guild."""


class BanManager(DatabaseChecker, Punisher):
    """
    A BanManager that manages guild bans.
    """

    __slots__ = ("bot",)

    def __init__(self, bot: commands.Bot):
        super().__init__(
            [
                {
                    "guild": "snowflake",
                    "member": "snowflake",
                    "reason": "string",
                    "timestamp": "snowflake",
                }
            ],
            ["bans"],
        )
        self.bot = bot

        self.add_event(self.on_database_connect)

    async def on_database_connect(self):
        self.bot.loop.create_task(self.__check_bans())

    async def get_banned_members(self) -> List[Dict[str, Any]]:
        """
        |coro|

        This function returns all the members that are supposed to be unbanned but are banned.

        :return: The list of unbanned members.
        :rtype: List[Dict[str, Any]]
        """

        return [
            x
            for x in await self.database.select(self.tables["bans"], [], fetchall=True)
            if x["timestamp"] <= datetime.utcnow().timestamp()
        ]

    async def __check_bans(self) -> None:
        """
        |coro|

        A loop that ensures that members are unbanned when they need to.

        :return: None
        :rtype: None
        """

        await self.bot.wait_until_ready()

        while not self.bot.is_closed():
            for banned_member in await self.get_banned_members():
                guild = self.bot.get_guild(banned_member["guild"])

                if guild is None:
                    continue

                user = await self.bot.fetch_user(banned_member["member"])

                if await self.unban(user, guild):
                    await self.call_event("on_unban", user, banned_member["reason"])

            await asyncio.sleep(300)

    async def punish(
        self, ctx: commands.Context, member: discord.Member, punishment: Punishment
    ) -> None:
        try:
            self.bot.loop.create_task(
                self.ban(
                    member,
                    punishment.punishment_reason,
                    punishment.punishment_time.total_seconds(),
                )
            )
        except discord.errors.Forbidden as e:
            raise e
        else:
            await self.call_event("on_punishment", ctx, member, punishment)

    @staticmethod
    async def get_ban(
        member: Union[discord.Member, discord.User], guild: discord.Guild
    ) -> Optional[discord.User]:
        """
        |coro|

        This function returns the user object of the member if he is banned from the guild.

        :param member: The banned member.
        :type member: discord.Member
        :param guild: The guild.
        :type guild: discord.Guild
        :return: The user object if found.
        :rtype: Optional[discord.User]
        """

        banned = await guild.bans()
        for x in banned:
            if x.user.id == member.id:
                return x.user

    async def unban(
        self, member: Union[discord.Member, discord.User], guild: discord.Guild = None
    ) -> bool:
        self._check_database()

        if isinstance(member, discord.User) and not guild:
            raise UnbanFailure("Cannot unban a discord.User without a guild.")

        guild = guild if guild is not None else member.guild
        await self.database.delete(
            self.tables["bans"], {"guild": guild.id, "member": member.id}
        )

        if user := await self.get_ban(member, guild):
            await guild.unban(user)
            return True

    async def __handle_unban(
        self, time_of_ban: Union[int, float], member: discord.Member, reason: str
    ) -> None:
        """
        |coro|

        A function that handles the member's unban that runs separately from the ban method so it wont be blocked.

        :param Union[int, float] time_of_ban: The time until the member's unban timestamp.
        :param discord.Member member: The member to unban.
        :param str reason: The reason of the mute.
        :return: None
        :rtype: None
        """

        await asyncio.sleep(time_of_ban)

        if await self.unban(member):
            await self.call_event("on_unban", member, reason)

    async def ban(
        self,
        member: discord.Member,
        reason: str = "No reason provided.",
        time_of_ban: Union[int, float] = 0,
    ) -> None:
        """
        |coro|

        Bans the member from the guild.

        :param member: The member to ban.
        :type member: discord.Member
        :param reason: The reason of the ban.
        :type reason: str
        :param time_of_ban: The time of ban.
        :type time_of_ban: Union[int, float]
        :return: None
        :rtype: None
        """

        self._check_database()

        await member.ban(reason=reason)

        if time_of_ban <= 0:
            return

        await self.database.insert(
            self.tables["bans"],
            {
                "guild": member.guild.id,
                "member": member.id,
                "reason": reason,
                "timestamp": datetime.utcnow().timestamp() + time_of_ban,
            },
        )

        self.bot.loop.create_task(self.__handle_unban(time_of_ban, member, reason))



from __future__ import annotations

import uuid
from datetime import datetime
from typing import List, TYPE_CHECKING, Optional, Dict, Union

from .base import DatabaseChecker
from .punishments import Punisher, get_relevant_punishment

if TYPE_CHECKING:
    from .punishments import Punishment
    import discord
    from discord.ext import commands


__all__ = ("PartialInfraction", "Infraction", "InfractionManager")


class PartialInfraction:
    """
    A partial infraction.
    """

    def __init__(
        self,
        member: discord.Member,
        infraction_id: str,
        reason: str,
        date_of_infraction: datetime,
    ):
        self.member = member
        self.infraction_id = infraction_id
        self.reason = reason
        self.date_of_infraction = date_of_infraction


class Infraction:
    """
    An infraction object.
    """

    def __init__(
        self, database, table: str, member: discord.Member, infraction_id: str
    ):
        self.member = member
        self.database = database
        self.table = table
        self.infraction_id = infraction_id

    def __str__(self):
        return f"<Infraction {self.infraction_id=}>"

    def __repr__(self):
        return f"<Infraction {self.member=}, {self.infraction_id=}>"

    @property
    def __checks(self) -> Dict[str, int]:
        return {
            "guild": self.member.guild.id,
            "member": self.member.id,
            "id": self.infraction_id,
        }

    async def datetime(self) -> Optional[datetime]:
        timestamp_data = await self.database.select(
            self.table, ["timestamp"], self.__checks
        )
        if timestamp_data:
            return datetime.utcfromtimestamp(timestamp_data["timestamp"])

    async def reason(self) -> Optional[str]:
        reason_data = await self.database.select(self.table, ["reason"], self.__checks)
        if reason_data:
            return reason_data["reason"]

    async def set_reason(self, new_reason: str) -> None:
        await self.database.update(self.table, {"reason": new_reason}, self.__checks)

    async def delete(self) -> PartialInfraction:
        partial = PartialInfraction(
            self.member, self.infraction_id, await self.reason(), await self.datetime()
        )
        await self.database.delete(self.table, self.__checks)
        return partial


class InfractionManager(DatabaseChecker, Punisher):
    def __init__(self, bot: commands.Bot):
        super().__init__(
            [
                {
                    "guild": "snowflake",
                    "member": "snowflake",
                    "timestamp": "snowflake",
                    "id": "string",
                    "reason": "string",
                }
            ],
            ["infractions"],
        )
        self.punishments = []
        self.bot = bot

    def add_punishments(self, punishments: List[Punishment]) -> None:
        self.punishments = punishments

    async def warn(
        self, ctx: commands.Context, member: discord.Member, reason: str
    ) -> Infraction:
        self._check_database()

        generated_id = str(uuid.uuid4())
        await self.database.insert(
            self.tables["infractions"],
            {
                "guild": member.guild.id,
                "member": member.id,
                "timestamp": datetime.utcnow().timestamp(),
                "id": generated_id,
                "reason": reason,
            },
        )

        if punishment := get_relevant_punishment(
            self.punishments, len(await self.get_infractions(member))
        ):
            await punishment.punishment_manager.punish(ctx, member, punishment)

        return Infraction(
            self.database, self.tables["infractions"], member, generated_id
        )

    async def punish(
        self, ctx: commands.Context, member: discord.Member, punishment: Punishment
    ) -> None:
        await self.warn(ctx, member, punishment.punishment_reason)
        await self.call_event("on_punishment", ctx, member, punishment)

    async def get_infractions(
        self,
        member: discord.Member,
        infraction_id: str = None,
        from_timestamp: Union[int, float] = 0,
    ) -> List[Infraction]:
        self._check_database()

        checks = {"guild": member.guild.id, "member": member.id}
        if infraction_id:
            checks["id"] = infraction_id

        warnings = await self.database.select(
            self.tables["infractions"], [], checks, fetchall=True
        )

        return [
            Infraction(
                self.database, self.tables["infractions"], member, infraction["id"]
            )
            for infraction in warnings
            if infraction["timestamp"] > from_timestamp
        ]



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from discord.ext import commands
    from .http import HTTPClient


__all__ = ("Interaction", "OptionType", "RespondType")


class RespondType(Enum):
    """Respond Types :)"""


class OptionType(Enum):
    SUB_COMMAND = 1
    SUB_COMMAND_GROUP = 2
    STRING = 3
    INTEGER = 4
    BOOLEAN = 5
    USER = 6
    CHANNEL = 7
    ROLE = 8
    MENTIONABLE = 9
    NUMBER = 10


class Interaction:
    """
    Represents a slash interaction.
    """

    __slots__ = ("bot", "http", "raw", "data", "args", "id", "token")

    def __init__(self, bot: commands.Bot, raw_interaction, http: HTTPClient):
        self.bot: commands.Bot = bot
        self.http = http
        self.raw = raw_interaction
        self.data = raw_interaction["d"]
        self.args = self.data["data"]
        self.id = self.data["id"]
        self.token = self.data["token"]

    @property
    def author(self):
        """Not done."""

    @property
    def guild(self):
        """Not done."""

    @property
    def channel(self):
        """Not done."""

    def __str__(self):
        return f"<{self.__class__.__name__} guild={self.guild}, author={self.author}>"

    def __repr__(self):
        return f"<{self.__str__()}, channel={self.channel}>"

    async def reply(self, text: str) -> None:
        """
        |coro|

        Reply to an interaction.

        :param text: The text to send.
        :type text: str
        :return: The message
        :rtype: None
        """

        url = f"https://discord.com/api/v9/interactions/{self.id}/{self.token}/callback"
        payload = {"type": 4, "data": {"content": text}}

        await self.http.request("POST", url, payload=payload)  # Message not done.



from .antispam import SpamDetectionGenerator, SpamManager
from .ban import BanManager
from .base import CogManager, questionnaire
from .birthday import BirthdayManager
from .commandhinter import CommandHinter, CommandResponseGenerator
from .convertors import TimeConvertor
from .database import DatabaseManager, create_mysql
from .economy import EconomyManager, EconomyAccount
from .fivem import FiveMServer
from .imaging import ImageManager, Backgrounds
from .infractions import InfractionManager
from .interactions.client import SlashManager
from .interactions.interaction import Interaction, OptionType
from .invitetracker import InviteTracker
from .kick import KickManager
from .leveling import LevelingManager
from .messagefilter import MessageFilter, MessageResponseGenerator
from .music.exceptions import *
from .music.music import *
from .music.player import Player
from .mute import MuteManager, AlreadyMuted
from .paginator import PageManager, generate_embeds, ButtonsPageManager
from .prefix import PrefixManager
from .punishments import Punishment
from .reactionroles import ReactionManager
from .spotify import SpotifyClient
from .template import TemplateManager
from .youtube import YoutubeClient

__title__ = "discordSuperUtils"
__version__ = "0.2.1"
__author__ = "Koyashie07 & Adam7100"
__license__ = "MIT"



from __future__ import annotations

import asyncio
from typing import Optional, TYPE_CHECKING, List, Iterable

from ..youtube import YoutubeClient

if TYPE_CHECKING:
    import discord


__all__ = ("Player",)

YOUTUBE = YoutubeClient()


class Player:
    """
    Represents a music player.
    """

    __slots__ = (
        "data",
        "title",
        "stream_url",
        "url",
        "start_timestamp",
        "last_pause_timestamp",
        "duration",
        "requester",
        "source",
        "autoplayed",
    )

    def __init__(self, requester: Optional[discord.Member], data):
        self.source = None
        self.data = data
        self.requester = requester
        self.title = data["videoDetails"]["title"]
        self.stream_url = data.get("url")
        self.url = "https://youtube.com/watch/?v=" + data["videoDetails"]["videoId"]

        self.autoplayed = False
        self.start_timestamp = 0
        self.last_pause_timestamp = 0

        duration = int(data["videoDetails"]["lengthSeconds"])
        self.duration = duration if duration != 0 else "LIVE"

    def __str__(self):
        return self.title

    @staticmethod
    def _get_stream_url(player: dict) -> str:
        """
        Returns the stream url of a player.

        :param dict player: The player.
        :return: The stream url.
        :rtype: str
        """

        stream_urls = [
            x
            for x in sorted(
                player["streamingData"]["adaptiveFormats"],
                key=lambda x: x.get("averageBitrate", 0),
                reverse=True,
            )
            if "audio" in x["mimeType"] and "opus" not in x["mimeType"]
        ]

        return player["streamingData"].get("hlsManifestUrl") or stream_urls[0]["url"]

    @classmethod
    async def make_multiple_players(
        cls, songs: Iterable[str], requester: Optional[discord.Member]
    ) -> List[Player]:
        """
        |coro|

        Returns a list of players from a iterable of queries.

        :param requester: The requester.
        :type requester: Optional[discord.Member]
        :param songs: The queries.
        :type songs: Iterable[str]
        :return: The list of created players.
        :rtype: List[Player]
        """

        tasks = [cls.fetch_song(song, playlist=False) for song in songs]

        songs = await asyncio.gather(*tasks)

        return [cls(requester, data=x[0]) for x in songs if x]

    @classmethod
    async def get_similar_videos(cls, video_id: str) -> List[Player]:
        """
        |coro|

        Creates similar videos related to the video id.

        :param str video_id: The video id
        :return: The list of similar players
        :rtype: List[Player]
        """

        similar_video = await YOUTUBE.get_similar_videos(video_id)
        players = await cls.make_players(
            f"https://youtube.com/watch/?v={similar_video[0]}", None, False
        )
        for player in players:
            player.autoplayed = True

        return players

    @staticmethod
    async def fetch_data(query: str, playlist: bool = True) -> List[dict]:
        """
        |coro|

        Fetches the youtube data of the query.

        :param bool playlist: Indicating if it should fetch playlists.
        :param query: The query.
        :type query: str
        :return: The youtube data.
        :rtype: Optional[dict]
        """

        return [
            x
            for x in await YOUTUBE.get_videos(
                await YOUTUBE.get_query_id(query), playlist
            )
            if "streamingData" in x
        ]

    @classmethod
    async def fetch_song(cls, query: str, playlist: bool = True) -> List[dict]:
        """
        |coro|

        Fetches the song's or playlist's data.
        Will return the first song in the playlist if playlist is False.

        :param query: The query.
        :type query: str
        :param playlist: A bool indicating if the function should fetch playlists or get the first video.
        :type playlist: bool
        :return: The list of songs.
        :rtype: List[dict]
        """

        data = await cls.fetch_data(query, playlist)
        if not data:
            return []

        players = []
        for player in data:
            player["url"] = cls._get_stream_url(player)
            players.append(
                {x: y for x, y in player.items() if x in ["url", "videoDetails"]}
            )

        return players

    @classmethod
    async def make_players(
        cls, query: str, requester: Optional[discord.Member], playlist: bool = True
    ) -> List[Player]:
        """
        |coro|

        Returns a list of players from the query.

        :param Optional[discord.Member] requester: The song requester.
        :param query: The query.
        :type query: str
        :param playlist: A bool indicating if the function should fetch playlists or get the first video.
        :type playlist: bool
        :return: The list of created players.
        :rtype: List[Player]
        """

        return [
            cls(requester, data=player)
            for player in await cls.fetch_song(query, playlist)
        ]



from __future__ import annotations

import asyncio
import inspect
from typing import (
    List,
    Any,
    Iterable,
    Optional,
    TYPE_CHECKING,
    Union,
    Tuple,
    Callable,
    Dict,
    Awaitable,
)

import aiomysql
import aiopg
import aiosqlite
import discord
from motor import motor_asyncio

if TYPE_CHECKING:
    from discord.ext import commands
    from .database import Database
    from datetime import timedelta


__all__ = (
    "COLUMN_TYPES",
    "DatabaseNotConnected",
    "InvalidGenerator",
    "get_generator_response",
    "maybe_coroutine",
    "generate_column_types",
    "questionnaire",
    "EventManager",
    "CogManager",
    "DatabaseChecker",
    "CacheBased",
)


COLUMN_TYPES = {
    motor_asyncio.AsyncIOMotorDatabase: None,  # mongo does not require any columns
    aiosqlite.core.Connection: {
        "snowflake": "INTEGER",
        "string": "TEXT",
        "number": "INTEGER",
        "smallnumber": "INTEGER",
    },
    aiopg.pool.Pool: {
        "snowflake": "bigint",
        "string": "character varying",
        "number": "integer",
        "smallnumber": "smallint",
    },
    aiomysql.pool.Pool: {
        "snowflake": "BIGINT",
        "string": "TEXT",
        "number": "INT",
        "smallnumber": "SMALLINT",
    },
}


class DatabaseNotConnected(Exception):
    """Raises an error when the user tries to use a method of a manager without a database connected to it."""


class CacheBased:
    """
    Represents a cache manager that manages member cache.
    """

    def __init__(self, bot: commands.Bot, wipe_cache_delay: timedelta):
        self.wipe_cache_delay = wipe_cache_delay
        self.bot = bot
        self._cache = {}

        asyncio.get_event_loop().create_task(self.__wipe_cache())

    async def __wipe_cache(self) -> None:
        """
        |coro|

        This function is responsible for wiping the member cache.

        :return: None
        :rtype: None
        """

        while not self.bot.is_closed():
            await asyncio.sleep(self.wipe_cache_delay.total_seconds())

            self._cache = {}


class InvalidGenerator(Exception):
    """
    Raises an exception when the user passes an invalid generator.
    """

    __slots__ = ("generator",)

    def __init__(self, generator):
        self.generator = generator
        super().__init__(
            f"Generator of type {type(self.generator)!r} is not supported."
        )


async def maybe_coroutine(function: Callable, *args, **kwargs) -> Any:
    """
    |coro|

    Returns the coroutine version of the function.

    :param function: The function to convert.
    :type function: Union[Awaitable, Callable]
    :param args: The arguments.
    :param kwargs: The key arguments:
    :return: The coroutine version of the function.
    :rtype: Awaitable
    """

    value = function(*args, **kwargs)

    if inspect.isawaitable(value):
        return await value

    return value


def get_generator_response(generator: Any, generator_type: Any, *args, **kwargs) -> Any:
    """
    Returns the generator response with the arguments.

    :param generator: The generator to get the response from.
    :type generator: Any
    :param generator_type: The generator type. (Should be same as the generator type.
    :type generator_type: Any
    :param args: The arguments of the generator.
    :param kwargs: The key arguments of the generator
    :return: The generator response.
    :rtype: Any
    """

    if inspect.isclass(generator) and issubclass(generator, generator_type):
        if inspect.ismethod(generator.generate):
            return generator.generate(*args, **kwargs)

        return generator().generate(*args, **kwargs)

    if isinstance(generator, generator_type):
        return generator.generate(*args, **kwargs)

    raise InvalidGenerator(generator)


def generate_column_types(
    types: Iterable[str], database_type: Any
) -> Optional[List[str]]:
    """
    Generates the column type names that are suitable for the database type.

    :param types: The column types.
    :type types: Iterable[str]
    :param database_type: The database type.
    :type database_type: Any
    :return: The suitable column types for the database types.
    :rtype: Optional[List[str]]
    """

    database_type_configuration = COLUMN_TYPES.get(database_type)

    if database_type_configuration is None:
        return

    return [database_type_configuration[x] for x in types]


async def questionnaire(
    ctx: commands.Context,
    questions: Iterable[Union[str, discord.Embed]],
    public: bool = False,
    timeout: Union[float, int] = 30,
    member: discord.Member = None,
) -> Tuple[List[str], bool]:
    """
    |coro|

    Questions the member using a "quiz" and returns the answers.
    The questionnaire can be used without a specific member and be public.
    If no member was passed and the questionnaire public argument is true, a ValueError will be raised.

    :raises: ValueError: The questionnaire is private and no member was provided.
    :param ctx: The context (where the questionnaire will ask the questions).
    :type ctx: commands.Context
    :param questions: The questions the questionnaire will ask.
    :type questions: Iterable[Union[str, discord.Embed]]
    :param public: A bool indicating if the questionnaire is public.
    :type public: bool
    :param timeout: The number of seconds until the questionnaire will stop and time out.
    :type timeout: Union[float, int]
    :param member: The member the questionnaire will get the answers from.
    :type member: discord.Member
    :return: The answers and a boolean indicating if the questionnaire timed out.
    :rtype: Tuple[List[str], bool]
    """

    answers = []
    timed_out = False

    if not public and not member:
        raise ValueError("The questionnaire is private and no member was provided.")

    def checks(msg):
        return (
            msg.channel == ctx.channel
            if public
            else msg.channel == ctx.channel and msg.author == member
        )

    for question in questions:
        if isinstance(question, str):
            await ctx.send(question)
        elif isinstance(question, discord.Embed):
            await ctx.send(embed=question)
        else:
            raise TypeError("Question must be of type 'str' or 'discord.Embed'.")

        try:
            message = await ctx.bot.wait_for("message", check=checks, timeout=timeout)
        except asyncio.TimeoutError:
            timed_out = True
            break

        answers.append(message.content)

    return answers, timed_out


class EventManager:
    """
    An event manager that manages events for managers.
    """

    def __init__(self):
        self.events = {}

    async def call_event(self, name: str, *args, **kwargs) -> None:
        """
        Calls the event name with the arguments

        :param name: The event name.
        :type name: str
        :param args: The arguments.
        :param kwargs: The key arguments.
        :return: None
        :rtype: None
        """

        if name in self.events:
            for event in self.events[name]:
                await event(*args, **kwargs)

    def event(self, name: str = None) -> Callable:
        """
        A decorator which adds an event listener.

        :param name: The event name.
        :type name: str
        :return: The inner function.
        :rtype: Callable
        """

        def inner(func):
            self.add_event(func, name)
            return func

        return inner

    def add_event(self, func: Callable, name: str = None) -> None:
        """
        Adds an event to the event dictionary.

        :param func: The event callback.
        :type func: Callable
        :param name: The event name.
        :type name: str
        :return: None
        :rtype: None
        :raises: TypeError: The listener isn't async.
        """

        name = func.__name__ if not name else name

        if not asyncio.iscoroutinefunction(func):
            raise TypeError("Listeners must be async.")

        if name in self.events:
            self.events[name].append(func)
        else:
            self.events[name] = [func]

    def remove_event(self, func: Callable, name: str = None) -> None:
        """
        Removes an event from the event dictionary.

        :param func: The event callback.
        :type func: Callable
        :param name: The event name.
        :type name: str
        :return: None
        :rtype: None
        """

        name = func.__name__ if not name else name

        if name in self.events:
            self.events[name].remove(func)


class CogManager:
    """
    A CogManager which helps the user use the managers inside discord cogs.
    """

    class Cog:
        """
        The internal Cog class.
        """

        def __init__(self, managers: List = None):
            listeners = {}
            managers = [] if managers is None else managers

            attribute_objects = [getattr(self, attr) for attr in dir(self)]

            for attr in attribute_objects:
                listener_type = getattr(attr, "_listener_type", None)
                if listener_type:
                    if listener_type in listeners:
                        listeners[listener_type].append(attr)
                    else:
                        listeners[listener_type] = [attr]

            managers = managers or [
                attr for attr in attribute_objects if type(attr) in listeners
            ]
            for event_type in listeners:
                for manager in managers:
                    for event in listeners[event_type]:
                        manager.add_event(event)

    @staticmethod
    def event(manager_type: Any) -> Callable:
        """
        Adds an event to the Cog event list.

        :param manager_type: The manager type of the event.
        :type manager_type: Any
        :rtype: Callable
        :return: The inner function.
        :raises: TypeError: The listener isn't async.
        """

        def decorator(func):
            if not inspect.iscoroutinefunction(func):
                raise TypeError("Listeners must be async.")

            func._listener_type = manager_type

            return func

        return decorator


class DatabaseChecker(EventManager):
    """
    A database checker which makes sure the database is connected to a manager and handles the table creation.
    """

    def __init__(
        self, tables_column_data: List[Dict[str, str]], table_identifiers: List[str]
    ):
        super().__init__()

        self.database = None
        self.table_identifiers = table_identifiers
        self.tables = {}
        self.tables_column_data = tables_column_data

    def _check_database(self, raise_error: bool = True) -> bool:
        """
        A function which checks if the database is connected.

        :param raise_error: A bool indicating if the function should raise an error if the database is not connected.
        :type raise_error: bool
        :rtype: bool
        :return: If the database is connected.
        :raises: DatabaseNotConnected: The database is not connected.
        """

        if not self.database:
            if raise_error:
                raise DatabaseNotConnected(
                    f"Database not connected."
                    f" Connect this manager to a database using 'connect_to_database'"
                )

            return False

        return True

    async def connect_to_database(self, database: Database, tables: List[str]) -> None:
        """
        Connects to the database.
        Calls on_database_connect when connected.

        :param database: The database to connect to.
        :type database: Database
        :param tables: The tables to create (incase they do not exist).
        :type tables: List[str]
        :rtype: None
        :return: None
        """

        for table, table_data, identifier in zip(
            tables, self.tables_column_data, self.table_identifiers
        ):
            types = generate_column_types(table_data.values(), type(database.database))

            await database.create_table(
                table, dict(zip(list(table_data), types)) if types else None, True
            )

            self.database = database
            self.tables[identifier] = table

        await self.call_event("on_database_connect")



from typing import Union, Any

import discord
from discord.ext import commands

from .base import DatabaseChecker


class PrefixManager(DatabaseChecker):
    def __init__(self, bot: commands.Bot, default_prefix: str, mentioned: bool = False):
        super().__init__([{"guild": "snowflake", "prefix": "string"}], ["prefixes"])
        self.default_prefix = default_prefix
        self.bot = bot
        self.mentioned = mentioned

        self.prefix_cache = {}
        bot.command_prefix = self.__get_prefix

    async def get_prefix(self, guild: Union[discord.Guild, Any]) -> str:
        if guild.id in self.prefix_cache:
            return self.prefix_cache[guild.id]

        prefix = await self.database.select(
            self.tables["prefixes"], ["prefix"], {"guild": guild.id}
        )
        prefix = prefix["prefix"] if prefix else self.default_prefix

        self.prefix_cache[guild.id] = prefix

        return prefix

    async def set_prefix(self, guild: discord.Guild, prefix: str) -> None:
        self.prefix_cache[guild.id] = prefix
        await self.database.updateorinsert(
            self.tables["prefixes"],
            {"prefix": prefix},
            {"guild": guild.id},
            {"guild": guild.id, "prefix": prefix},
        )

    async def __get_prefix(self, bot, message: discord.Message) -> str:
        self._check_database()

        if not message.guild:
            return (
                commands.when_mentioned_or(self.default_prefix)(bot, message)
                if self.mentioned
                else self.default_prefix
            )

        prefix = await self.get_prefix(message.guild)

        return (
            commands.when_mentioned_or(prefix)(bot, message)
            if self.mentioned
            else prefix
        )



from .base import DatabaseChecker
from .paginator import EmojiError


class ReactionManager(DatabaseChecker):
    def __init__(self, bot):
        super().__init__(
            [
                {
                    "guild": "snowflake",
                    "message": "snowflake",
                    "role": "snowflake",
                    "emoji": "string",
                    "remove_on_reaction": "smallnumber",
                }
            ],
            ["reaction_roles"],
        )

        self.bot = bot
        self.add_event(self.on_database_connect)

    async def on_database_connect(self):
        self.bot.add_listener(self.__handle_reactions, "on_raw_reaction_add")
        self.bot.add_listener(self.__handle_reactions, "on_raw_reaction_remove")

    @staticmethod
    def get_emoji_sql(emoji):
        if not emoji.is_custom_emoji():
            return str(emoji)

        emoji_string = f"<:{emoji.name}:{emoji.id}>"
        if emoji.animated:
            emoji_string = emoji_string[:1] + "a" + emoji_string[1:]

        return emoji_string

    async def __handle_reactions(self, payload):
        self._check_database()

        if payload.user_id == self.bot.user.id:
            return

        database_checks = {
            "guild": payload.guild_id,
            "message": payload.message_id,
            "emoji": self.get_emoji_sql(payload.emoji),
        }

        reaction_role_data = await self.database.select(
            self.tables["reaction_roles"], self.tables_column_data[0], database_checks
        )

        if not reaction_role_data:
            return

        guild = self.bot.get_guild(payload.guild_id)
        role = guild.get_role(reaction_role_data["role"])
        channel = guild.get_channel(payload.channel_id)
        message = await channel.fetch_message(payload.message_id)
        emoji = self.get_emoji_sql(payload.emoji)

        if emoji == reaction_role_data["emoji"]:
            member = (
                payload.member if payload.member else guild.get_member(payload.user_id)
            )

            if role is None:
                await self.call_event(
                    "on_reaction_event", guild, channel, message, member, emoji
                )

            else:
                if role not in member.roles:
                    await member.add_roles(role)
                elif reaction_role_data["remove_on_reaction"] == 1:
                    await member.remove_roles(role)

    async def create_reaction(
        self, guild, message, role, emoji, remove_on_reaction: int
    ):
        self._check_database()

        await self.database.insertifnotexists(
            self.tables["reaction_roles"],
            dict(
                zip(
                    self.tables_column_data[0],
                    [
                        guild.id,
                        message.id,
                        role.id if role is not None else role,
                        emoji,
                        int(remove_on_reaction),
                    ],
                )
            ),
            {"guild": guild.id, "message": message.id, "emoji": emoji},
        )

        if len(emoji) > 1:
            emoji = self.bot.get_emoji(emoji)

        try:
            await message.add_reaction(emoji)
        except Exception:
            raise EmojiError("Cannot add reaction to message.")

    async def delete_reaction(self, guild, message, emoji):
        await self.database.delete(
            self.tables["reaction_roles"],
            {"guild": guild.id, "message": message.id, "emoji": emoji},
        )

    async def get_reactions(self, guild=None):
        return await self.database.select(
            self.tables["reaction_roles"],
            [],
            {"guild": guild.id} if guild else {},
            True,
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import timedelta
from typing import Optional, List, TYPE_CHECKING

if TYPE_CHECKING:
    import discord
    from discord.ext import commands


class Punishment:
    """
    A punishment class that is used for punishing members.
    """

    def __init__(
        self,
        punishment_manager,
        punish_after: int = 3,
        punishment_reason: str = "No reason specified.",
        punishment_time: timedelta = timedelta(days=1),
    ):
        self.punishment_manager = punishment_manager
        self.punish_after = punish_after
        self.punishment_reason = punishment_reason
        self.punishment_time = punishment_time

        if not issubclass(type(punishment_manager), Punisher):
            raise TypeError(
                f"Manager of type '{type(punishment_manager)} is not supported.'"
            )


def get_relevant_punishment(
    punishments: List[Punishment], punish_count: int
) -> Optional[Punishment]:
    """
    Returns the punishment that is suitable for the punish count.

    :param punishments: The punishments to pick from.
    :type punishments: List[Punishment]
    :param punish_count: The punishment count.
    :type punish_count: int
    :rtype: Optional[Punishment]
    :return: The suitable punishment.
    """

    return {x.punish_after: x for x in punishments}.get(punish_count)


class Punisher(ABC):
    @abstractmethod
    async def punish(
        self, ctx: commands.Context, member: discord.Member, punishment: Punishment
    ) -> None:
        """
        The manager's punish function.

        :param ctx: The context of the punishments.
        :type ctx: commands.Context
        :param member: The member to punish.
        :type member: discord.Member
        :param punishment: The punishment to punish the member with.
        :type punishment: Punishment
        :rtype: None
        :return: None
        """



from __future__ import annotations

from typing import Dict, Any

import aiohttp


__all__ = ("HTTPClient",)


class HTTPClient:
    """
    Represents a HTTP client that handles interactions.
    """

    __slots__ = ("bot_token", "session")

    def __init__(self, bot_token: str, session: aiohttp.ClientSession = None):
        self.bot_token = bot_token
        self.session = session or aiohttp.ClientSession()

    async def request(
        self,
        request_type: str,
        url: str,
        payload: Dict[str, Any] = None,
        headers: Dict[str, Any] = None,
    ) -> str:
        """
        |coro|

        The website request method of the HTTP client.

        :param request_type: The request type.
        :type request_type: str
        :param url: The url.
        :type url: str
        :param payload: The payload.
        :type payload: Dict[str, Any]
        :param headers: The headers.
        :type headers: Dict[str, Any]
        :return: The response.
        :rtype: str
        """

        r = await self.session.request(
            request_type.upper(), url, json=payload or {}, headers=headers or {}
        )
        return await r.text()

    async def add_slash_command(
        self, payload: Dict[str, Any], application_id: int
    ) -> None:
        """
        |coro|

        The slash command creator method.

        :param payload: The payload of the creation.
        :type payload: Dict[str, Any]
        :param application_id: The application id of the bot (user id)
        :type application_id: int
        :rtype: None
        :return: None
        """

        url = f"https://discord.com/api/v9/applications/{application_id}/commands"
        headers = {
            "Authorization": self.bot_token,
        }

        await self.request("POST", url, payload, headers)



from setuptools import setup

f = open("README.md", "r")
README = f.read()

setup(
    name="discordSuperUtils",
    packages=["discordSuperUtils"],
    package_data={
        "discordSuperUtils.assets": ["*"],
        "": ["*.png", "*.ttf"],
        "discordSuperUtils.interactions": ["*"],
        "discordSuperUtils.music": ["*"],
    },
    include_package_data=True,
    version="0.2.1",
    license="MIT",
    description="Discord Bot Development made easy!",
    long_description=README,
    long_description_content_type="text/markdown",
    author="koyashie07 and adam7100",
    url="https://github.com/discordsuperutils/discord-super-utils",
    download_url="https://github.com/discordsuperutils/discord-super-utils/archive/refs/tags/v0.2.1.tar.gz",
    keywords=[
        "discord",
        "easy",
        "discord.py",
        "music",
        "download",
        "links",
        "images",
        "videos",
        "audio",
        "bot",
        "paginator",
        "economy",
        "reaction",
        "reaction roles",
        "database",
        "database manager",
    ],
    install_requires=[
        "discord.py",
        "Pillow",
        "requests",
        "spotipy",
        "aiosqlite",
        "motor",
        "aiopg",
        "aiomysql",
        "discord_components",
        "pytz",
    ],
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 4 - Beta",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        # Pick your license as you wish
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
)



import asyncio
import re
from typing import Dict, Any, List, Optional
from urllib import parse

import aiohttp

__all__ = ("YoutubeClient",)


class YoutubeClient:
    """
    Represents a Youtube client that fetches music.
    """

    __slots__ = ("session", "timeout")

    # This access key is not private, and is used in ALL youtube API requests from the website (from any user).
    ACCESS_KEY = "AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8"
    BASE_URL = "https://www.youtube.com/youtubei/v1"
    CONTEXT = (
        b"{'context': {'client': {'clientName': 'ANDROID','clientVersion': '16.20'}}}"
    )
    HEADERS = {
        "Content-Type": "application/json",
        "User-Agent": "Mozilla/5.0",
        "accept-language": "en-US,en",
    }

    def __init__(self, session: aiohttp.ClientSession = None, timeout: int = 10):
        self.timeout = timeout
        self.session = session or asyncio.get_event_loop().run_until_complete(
            self._make_session()
        )

    async def _make_session(self) -> aiohttp.ClientSession:
        """
        |coro|

        Makes an aiohttp session.

        :return: The session.
        :rtype: aiohttp.ClientSession
        """

        return aiohttp.ClientSession(
            timeout=aiohttp.ClientTimeout(total=self.timeout, connect=3)
        )

    async def request(
        self,
        url: str,
        query: Dict[str, Any],
        headers: Dict[str, str] = None,
        payload: bytes = CONTEXT,
    ) -> Optional[aiohttp.ClientResponse]:
        """
        |coro|

        Makes a request to the url with the specified arguments.

        :param str url: The url.
        :param Dict[str, Any] query: The query values.
        :param Dict[str, Any] headers: The request headers.
        :param bytes payload: The payload to send.
        :return: The request object.
        :rtype: aiohttp.ClientResponse
        """

        headers = headers or self.HEADERS

        try:
            return await self.session.post(
                f"{url}?{parse.urlencode(query)}", data=payload, headers=headers
            )
        except asyncio.exceptions.TimeoutError:
            return None

    async def get_query_id(self, query: str) -> Optional[str]:
        """
        |coro|

        Gets the query video or playlist id.

        :param str query: The query.
        :return: The video or playlist id is applicable.
        :rtype: Optional[str]
        """

        query_arguments = dict(parse.parse_qsl(parse.urlparse(query).query))

        result_id = query_arguments.get("list") or query_arguments.get("v")
        if result_id:
            return result_id

        search = await self.search(query)
        return search[0] if search else None

    async def search(self, query: str) -> List[str]:
        """
        |coro|

        Returns the video ids found from the query.

        :param str query: The query.
        :return: The video ids.
        :rtype: List[str]
        """

        # Not using the youtube API as it is inconsistent.

        try:
            r = await self.session.get(
                f"https://www.youtube.com/results?search_query={query}"
            )
        except asyncio.exceptions.TimeoutError:
            return []

        return re.findall(r"watch\?v=(\S{11})", await r.text())

    async def get_similar_videos(self, video_id: str) -> List[str]:
        """
        |coro|

        Returns similar videos to the video id.
        Simply returns the suggestions in the suggestions section when watching a video.

        :param str video_id: The video id.
        :return: The list of similar videos.
        :rtype: List[str]
        """

        query = {
            "key": self.ACCESS_KEY,
            "contentCheckOk": True,
            "racyCheckOk": True,
            "videoId": video_id,
        }

        r = await self.request(f"{self.BASE_URL}/next", query=query)
        if not r:
            return []

        r_json = await r.json()

        suggestions = r_json["contents"]["singleColumnWatchNextResults"]["results"][
            "results"
        ]["contents"][1]

        return [
            x["gridVideoRenderer"]["videoId"]
            for x in suggestions["shelfRenderer"]["content"]["horizontalListRenderer"][
                "items"
            ]
            if "gridVideoRenderer" in x
        ]

    async def get_playlist_videos(self, playlist_id: str) -> List[str]:
        """
        |coro|

        Returns the video IDs in the playlist.

        :param str playlist_id: The playlist id.
        :return: The video IDs.
        :rtype: List[str]
        """

        query = {
            "key": self.ACCESS_KEY,
            "contentCheckOk": True,
            "racyCheckOk": True,
            "playlistId": playlist_id,
        }

        r = await self.request(f"{self.BASE_URL}/next", query=query)
        if not r:
            return []

        r_json = await r.json()

        # Youtube api lol
        return [
            x["playlistPanelVideoRenderer"]["navigationEndpoint"]["watchEndpoint"][
                "videoId"
            ]
            for x in r_json["contents"]["singleColumnWatchNextResults"]["playlist"][
                "playlist"
            ]["contents"]
        ]

    async def get_videos(
        self, video_id: str, playlist: bool = True
    ) -> List[Dict[str, Any]]:
        """
        |coro|

        Returns the videos data including title, duration, stream urls.

        :param bool playlist: Fetches playlists if True.
        :param str video_id: The video or playlist id.
        :return List[Dict[str, Any]]: The video data.
        """

        if not video_id:
            return []

        if len(video_id) == 11:  # Video
            queries = [
                {
                    "key": self.ACCESS_KEY,
                    "videoId": video_id,
                    "contentCheckOk": True,
                    "racyCheckOk": True,
                }
            ]

        else:  # Playlist
            queries = [
                {
                    "key": self.ACCESS_KEY,
                    "videoId": playlist_video_id,
                    "contentCheckOk": True,
                    "racyCheckOk": True,
                }
                for playlist_video_id in await self.get_playlist_videos(video_id)
            ]

        queries = queries[:1] if not playlist else queries

        requests = await asyncio.gather(
            *[self.request(f"{self.BASE_URL}/player", query=query) for query in queries]
        )
        try:
            return [await r.json() if r else {} for r in requests]
        except asyncio.exceptions.TimeoutError:
            return []



from __future__ import annotations

import inspect
from typing import Dict, Any, List, Callable

from discord.ext import commands

from .interaction import Interaction, OptionType
from .http import HTTPClient
from ..base import EventManager, maybe_coroutine

# discord.ext.commands is not type hinted in order to use the converters.

__all__ = ("SlashManager",)


class SlashManager(EventManager):
    """
    Represents a SlashManager that manages slash commands.
    """

    __slots__ = ("bot", "http", "commands")

    def __init__(self, bot: commands.Bot):
        super().__init__()
        self.bot = bot
        self.http = None
        self.commands = {}

        self.bot.add_listener(self.__dispatch_event, "on_socket_response")

        self.bot.loop.create_task(self.__initialize_http())

    async def __initialize_http(self) -> None:
        """
        |coro|

        Sets the SlashManager's http object when the bot is ready.

        :return: None
        :rtype: None
        """

        await self.bot.wait_until_ready()

        self.http = HTTPClient("Bot " + self.bot.http.token)

    async def process_command(self, interaction: Interaction) -> None:
        """
        |coro|

        Processes the command.

        :param interaction: The interaction to process the command from
        :type interaction: Interaction
        :return: None
        :rtype: None
        """

        converters = {
            6: [self.bot.get_user, self.bot.fetch_user],
            7: [interaction.guild and interaction.guild.get_channel],
            8: [interaction.guild and interaction.guild.get_role],
        }

        data = interaction.data["data"]
        name = data["name"]

        user_arguments = []

        if "options" in data:
            for argument in data["options"]:
                arg_type = argument["type"]
                if arg_type not in converters:
                    user_arguments.append(argument["value"])
                    continue

                result_fetch = argument["value"]
                for converter in converters[arg_type]:
                    convert_result = await maybe_coroutine(
                        converter, int(argument["value"])
                    )

                    if convert_result:
                        result_fetch = convert_result
                        break

                user_arguments.append(result_fetch)

        args = [interaction] + user_arguments
        await self.commands[name](*args)

    async def __dispatch_event(self, event: Dict[str, Any]) -> None:
        """
        |coro|

        Adds the on_interaction event to the SlashManager
        This function is the callback of on_socket_response.

        :param event: The socket response.
        :type event: Dict[str, Any]
        :return: None
        :rtype: None
        """

        if event["t"] == "INTERACTION_CREATE":
            interaction = Interaction(self.bot, event, self.http)

            await self.call_event("on_interaction", interaction)
            await self.process_command(interaction)

    async def add_global_command(self, payload: Dict[str, Any]) -> None:
        """
        |coro|

        Adds a global command from the payload.

        :param payload: The command options.
        :type payload: Dict[str, Any]
        :return: None
        :rtype: None
        """

        await self.bot.wait_until_ready()

        await self.http.add_slash_command(payload, self.bot.user.id)

    @staticmethod
    def parameters_to_dict(parameters) -> List[Dict[str, Any]]:
        """
        |coro|

        Converts signature parameters to their slash command JSON form.

        :param parameters: The parameters
        :return: The parameters JSON form.
        :rtype: List[Dict[str, Any]]
        """

        result_parameters = []

        for parameter in parameters:
            annotation = parameter.annotation

            result_parameters.append(
                {
                    "name": parameter.name,
                    "type": annotation.value
                    if isinstance(annotation, OptionType)
                    else 3,
                    "description": "-",
                    "required": True,
                }
            )

        return result_parameters

    def command(
        self, name: str = None, description: str = "No description"
    ) -> Callable:
        """
        The command initializer decorator.

        :param name: The command name.
        :type name: str
        :param description: The command description
        :type description: str
        :return: The inner function.
        :rtype: Callable
        """

        def inner(func):
            command_name = name or func.__name__

            sig = inspect.signature(func)

            command_initializer = {
                "name": command_name,
                "type": OptionType.SUB_COMMAND.value,
                "description": description,
                "options": self.parameters_to_dict(list(sig.parameters.values())[1:]),
            }

            self.commands[command_name] = func
            self.bot.loop.create_task(self.add_global_command(command_initializer))

        return inner



import aiohttp
import aiohttp.client_exceptions


class ServerNotFound(Exception):
    """Raises an error when a server is invalid or offline."""


class FiveMPlayer:
    def __init__(self, player_id, identifiers, name, ping):
        self.player_id = player_id
        self.identifiers = identifiers
        self.name = name
        self.ping = ping

    def __str__(self):
        return f"<FiveM Player {self.player_id=}>"

    def __repr__(self):
        return f"<FiveM Player {self.name=}, {self.player_id=}, {self.identifiers=}, {self.ping=}>"

    @classmethod
    def fetch(cls, player_dict):
        identifiers = dict([x.split(":") for x in player_dict["identifiers"]])
        return cls(
            player_dict["id"], identifiers, player_dict["name"], player_dict["ping"]
        )


class FiveMServer:
    def __init__(self, ip, resources, players, name, variables):
        self.ip = ip
        self.resources = resources
        self.players = players
        self.name = name
        self.variables = variables

    def __str__(self):
        return f"<FiveM Server {self.name=}>"

    def __repr__(self):
        return (
            f"<FiveM Server {self.ip=},"
            f" {self.name=},"
            f" {self.players=},"
            f" {self.resources=},"
            f" {self.variables=}>"
        )

    @classmethod
    async def fetch(cls, ip):
        base_address = "http://" + ip + "/"

        async with aiohttp.ClientSession() as session:
            try:
                await session.get(base_address)  # Server status check
            except (
                aiohttp.client_exceptions.ClientConnectorError,
                aiohttp.client_exceptions.InvalidURL,
            ):
                raise ServerNotFound(f"Server '{ip}' is invalid or offline.")

            players_request = await session.get(base_address + "players.json")
            info_request = await session.get(base_address + "info.json")
            dynamic_info_request = await session.get(base_address + "dynamic.json")

            info = await info_request.json(content_type=None)
            players = await players_request.json(content_type=None)
            dynamic = await dynamic_info_request.json(content_type=None)
            # This is still included in the session because parsing the json outside of it sometimes doesnt work
            # and block the program (?) :(

        return cls(
            ip,
            info["resources"],
            [FiveMPlayer.fetch(player) for player in players],
            dynamic["hostname"],
            info["vars"],
        )
